
/**
 * @addtogroup cp_hashtable
 */
/** @{ */

/**
 * @file
 * Implementation of generic synchronized cp_hashtable. 
 *
 * The elements are stored in cp_list_entry objects.
 *
 * @copydoc collection
 */
/* ----------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <pthread.h>
#include <errno.h>
#include "hashtable.h"
#include "linked_list.h"
#include "log.h"
#include "common.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifndef CP_HASHTABLE_MULTIPLE_VALUES
#define CP_HASHTABLE_MULTIPLE_VALUES 1
#endif

static int table_sizes[] = 
    {
          11,      23,      47,      79,      97, 
         157,     197,     299,     397,     599, 
         797,    1297,    1597,    2499,    3199, 
        4799,    6397,    9599,   12799,   19199, 
       25597,   38399,   51199,   76799,  102397, 
      153599,  204797,  306797,  409597,  614399, 
      819199, 1288799, 1638397, 2457599, 3276799 
    };

static int table_sizes_len = 35;


long cp_hashtable_choose_size(long size_request)
{
    long new_size;

    if (table_sizes[table_sizes_len - 1] < size_request)
    {
        for (new_size = table_sizes[table_sizes_len - 1]; 
             new_size < size_request; 
             new_size = new_size * 2 + 1);
    }
    else
    {
        int min = -1;
        int max = table_sizes_len - 1;
        int pos;

        while (max > min + 1)
        {
            pos = (max + min + 1) / 2;
            if (table_sizes[pos] < size_request)
                min = pos;
            else
                max = pos;
        }

        new_size = table_sizes[max];
    }
         
    return new_size;
}

cp_hashtable *cp_hashtable_create(long initial_size, 
                                  cp_hashfunction hash_fn, 
                                  cp_compare_fn compare_fn)
{
    return cp_hashtable_create_by_option(0, initial_size, hash_fn, 
                                         compare_fn, NULL, NULL, NULL, NULL);
}


cp_hashtable *
    cp_hashtable_create_copy_mode(long initial_size, 
                                  cp_hashfunction hash_fn, 
                                  cp_compare_fn compare_fn, 
                                  cp_copy_fn copy_key, 
                                  cp_destructor_fn free_key,
                                  cp_copy_fn copy_value,
                                  cp_destructor_fn free_value)
{
    return cp_hashtable_create_by_option(COLLECTION_MODE_DEEP | 
                                         COLLECTION_MODE_COPY, 
                                         initial_size, 
                                         hash_fn, 
                                         compare_fn, 
                                         copy_key, 
                                         free_key,
                                         copy_value,
                                         free_value);
}

cp_hashtable *
    cp_hashtable_create_by_option(int mode, long initial_size, 
                                  cp_hashfunction hash_fn, 
                                  cp_compare_fn compare_fn, 
                                   cp_copy_fn copy_key, 
                                  cp_destructor_fn free_key,
                                  cp_copy_fn copy_value,
                                  cp_destructor_fn free_value)
{
    cp_hashtable *table;
    
    if (initial_size < 0)
    {
        errno = EINVAL;
        return NULL;
    }
    
    table = (cp_hashtable *) calloc(1, sizeof(cp_hashtable));
    if (table == NULL)
    {
        errno = ENOMEM;
        return NULL;
    }

    table->table_size = cp_hashtable_choose_size(initial_size);
    table->items = 0;
    table->table = (cp_hashtable_entry **) calloc(table->table_size, sizeof(cp_hashtable_entry *));
    if (table->table == NULL) 
    {
        errno = ENOMEM;
        return NULL;
    }

    table->hash_fn = hash_fn;
    table->compare_fn = compare_fn;
    table->copy_key = copy_key;
    table->copy_value = copy_value;
    table->free_key = free_key;
    table->free_value = free_value;

    table->mode = mode;

    table->min_size = initial_size;
    table->fill_factor_min = CP_HASHTABLE_DEFAULT_MIN_FILL_FACTOR;
    table->fill_factor_max = CP_HASHTABLE_DEFAULT_MAX_FILL_FACTOR;

    table->resizing = 0;

    table->lock = malloc(sizeof(pthread_rwlock_t));
    if (table->lock == NULL)
    {
        free(table->table);
        free(table);
        errno = ENOMEM;
        return NULL;
    }
        
    pthread_rwlock_init(table->lock, NULL);

    return table;
}

cp_hashtable_entry *cp_hashtable_create_entry(cp_hashtable *table, int mode, void *key, void *value, long cp_hashcode)
{
    cp_hashtable_entry *entry;

    entry = (cp_hashtable_entry *) malloc(sizeof(cp_hashtable_entry));
    if (entry == NULL) 
    {
        errno = ENOMEM;
        return NULL;
    }

    if (mode & COLLECTION_MODE_COPY) 
    {
        entry->key = table->copy_key ? (*table->copy_key)(key) : key;
        entry->value = table->copy_value ? (*table->copy_value)(value) :value;
    } 
    else 
    {
        entry->key = key;
        entry->value = value;    
    }

    entry->cp_hashcode = cp_hashcode;
    entry->next = NULL;

    return entry;
}

    
void cp_hashtable_destroy(cp_hashtable *table)
{
    long i;
    cp_hashtable_entry *entry, *next;
    

    if (table->resizing) 
    {
        pthread_cond_t resize;
        pthread_mutex_t resize_mutex;
        struct timespec delay;

        pthread_mutex_init(&resize_mutex, NULL);
        pthread_cond_init(&resize, NULL);
        delay.tv_sec = 0;
        delay.tv_nsec = 1e5;
        pthread_mutex_lock(&resize_mutex);
        while (table->resizing)
            pthread_cond_timedwait(&resize, &resize_mutex, &delay);
        pthread_mutex_unlock(&resize_mutex);
        pthread_mutex_destroy(&resize_mutex);
        pthread_cond_destroy(&resize);
    }
        
    for (i = 0; i < table->table_size; i++) 
    {
        entry = table->table[i];
        while (entry != NULL) 
        {
            next = entry->next;
            if (table->mode & COLLECTION_MODE_DEEP) 
            {
                if (table->free_key)
                    (*table->free_key)(entry->key);
                if (table->free_value)
                    (*table->free_value)(entry->value);
            }
            free(entry);
            entry = next;
        }
    }
        
    free(table->table);
    pthread_rwlock_destroy(table->lock);
    free(table->lock);
    free(table);
}

void cp_hashtable_destroy_deep(cp_hashtable *table)
{
    cp_hashtable_set_mode(table, COLLECTION_MODE_DEEP);
    cp_hashtable_destroy_custom(table, table->free_key, table->free_value);
}

void cp_hashtable_destroy_custom(cp_hashtable *table, cp_destructor_fn dk, cp_destructor_fn dv)
{
    long i;
    cp_hashtable_entry *entry, *next;

    if (table->resizing) 
    {
        pthread_cond_t resize;
        pthread_mutex_t resize_mutex;
        struct timespec delay;

        pthread_mutex_init(&resize_mutex, NULL);
        pthread_cond_init(&resize, NULL);
        delay.tv_sec = 0;
        delay.tv_nsec = 1e5;
        pthread_mutex_lock(&resize_mutex);
        while (table->resizing) 
            pthread_cond_timedwait(&resize, &resize_mutex, &delay);

        pthread_mutex_unlock(&resize_mutex);
        pthread_mutex_destroy(&resize_mutex);
        pthread_cond_destroy(&resize);
    }

    for (i = 0; i < table->table_size; i++) 
    {
        entry = table->table[i];
        while (entry != NULL) 
        {
            next = entry->next;
            if (dk) (*dk)(entry->key);
            if (dv) (*dv)(entry->value);
            free(entry);
            entry = next;
        }
    }
        
    free(table->table);
    pthread_rwlock_destroy(table->lock);
    free(table->lock);
    free(table);
}


void cp_hashtable_destroy_shallow(cp_hashtable *table)
{
    cp_hashtable_unset_mode(table, COLLECTION_MODE_DEEP);
    cp_hashtable_destroy(table);
}

int cp_hashtable_lock(cp_hashtable *table, int type)
{
    int rc;

    switch (type)
    {
        case COLLECTION_LOCK_READ:
            rc = pthread_rwlock_rdlock(table->lock);
            break;

        case COLLECTION_LOCK_WRITE:
            rc = pthread_rwlock_wrlock(table->lock);
            break;

        case COLLECTION_LOCK_NONE:
            rc = 0;
            break;

        default:
            rc = EINVAL;
            break;
    }

    return rc;
}
    

int cp_hashtable_unlock(cp_hashtable *table)
{
    return pthread_rwlock_unlock(table->lock);
}


void cp_hashtable_set_mode(cp_hashtable *table, int mode)
{
    int syncbit = table->mode & COLLECTION_MODE_NOSYNC;
    if (!syncbit)
        cp_hashtable_wrlock(table);
    table->mode |= mode;
    if (!syncbit)
        cp_hashtable_unlock(table);
}


void cp_hashtable_unset_mode(cp_hashtable *table, int mode)
{
    int syncbit = table->mode & COLLECTION_MODE_NOSYNC;
    if (!syncbit)
        cp_hashtable_wrlock(table);
    table->mode &= table->mode ^ mode;
    if (!syncbit)
        cp_hashtable_unlock(table);
}

/**
 * Retrieves the value by key from normal or resizing table.
 *
 * @param table the object
 * @param key    Key to search for.
 * @param code   Hash code of the Key (saves recalculating it)
 * @param option operation mode
 * @param resize 0: search in the normal table, 1: search in the resize table
 * @retval value of the entry with key.
 * @retval NULL otherwise (no entry with given key)
 */
void *lookup_internal(cp_hashtable *table, void *key, long code, int option, int resize)
{
    cp_hashtable_entry *entry;
    void *ret = NULL;

    entry = resize ? table->resize_table[code % table->resize_len]
                   : table->table[code % table->table_size];

    while (entry != NULL && (*table->compare_fn)(key, entry->key)) 
        entry = entry->next;
    
    if (entry)
    {
#if CP_HASHTABLE_MULTIPLE_VALUES
        if (option & COLLECTION_MODE_MULTIPLE_VALUES)
        {
            cp_list *l = 
                cp_list_create_view(table->mode, 
                                    NULL, 
                                    table->copy_value,
                                    table->free_value,
                                    table->lock);
            cp_list_insert(l, entry->value);
            entry = entry->next;
            while (entry != NULL) 
            {
                if ((*table->compare_fn)(key, entry->key) == 0) 
                    cp_list_append(l, entry->value);

                entry = entry->next;
            }

            ret = l;
        }
        else 
#endif
        ret = entry->value;
    }

    return ret;
}

/**
 * Retrieves the value by key.
 *
 * @param table the object
 * @param key Key to search for.
 * @param option operation mode
 * @retval value of the entry with key.
 * @retval NULL otherwise (no entry with given key or key == NULL)
 */
void *cp_hashtable_get_by_option(cp_hashtable *table, void *key, int option)
{
    long code;
    void *ret;
    int syncbit;

    if (table == NULL || key == NULL)
    {
        errno = EINVAL;
        return NULL;
    }

    syncbit = table->mode & COLLECTION_MODE_NOSYNC;
    if (!syncbit) cp_hashtable_rdlock(table);

    code = abs((*table->hash_fn)(key));
    ret = lookup_internal(table, key, code, option, 0);
    /* when resizing, search also there */
    if (ret == NULL && table->resizing) 
        ret = lookup_internal(table, key, code, option, 1);

    if (!syncbit) cp_hashtable_unlock(table);

    return ret;
}

/**
 * \<\<Thread\>\> resize a cp_hashtable.
 *
 * This cp_thread does a background resize of the table.
 * It creates a new table and moves the entries to the new table.
 * @post All items moved to the new table which replaces the old table
 *       The old table is destroyed.
 * @note The cp_thread locks and unlocks the table for each item.
 * This creates some overhead, but ensures that the table can still be used
 * during resizing.
 */
void *cp_hashtable_resize_thread(void *tbl)
{
    long i, old_size, index;
    long new_size;
    cp_hashtable_entry **new_table, **old_table;
    cp_hashtable_entry **insert, *entry;
    cp_hashtable *table = (cp_hashtable *) tbl;

    new_size = table->table_size;
    new_table = table->table;

    old_size = table->resize_len;
    old_table = table->resize_table;

#ifdef __TRACE__
    DEBUGMSG("resize %d to %d - resize thread starting\n", old_size, new_size);
#endif

    /* copy old table into new table, bucket by bucket */
    for (i = 0; i < old_size; i++) 
    {
//        if (!(table->mode & COLLECTION_MODE_NOSYNC))
        cp_hashtable_wrlock(table);
        entry = old_table[i];
        while (entry != NULL) 
        {
            index = entry->cp_hashcode % new_size;
            insert = &new_table[index];
            while (*insert != NULL)
                insert = &(*insert)->next;

            *insert = entry;
            entry = entry->next;
            (*insert)->next = NULL;
        }
        old_table[i] = NULL;

//        if (!(table->mode & COLLECTION_MODE_NOSYNC))
           cp_hashtable_unlock(table);
    }

    /* cleanup */
//    if (!(table->mode & COLLECTION_MODE_NOSYNC)) 
    cp_hashtable_wrlock(table);
    free(table->resize_table); table->resize_table = NULL;
    table->resizing = 0;
//    if (!(table->mode & COLLECTION_MODE_NOSYNC)) 
    cp_hashtable_unlock(table);

#ifdef __TRACE__
    DEBUGMSG("resize %d to %d - done\n", old_size, new_size);
#endif
    return NULL;
}

void cp_hashtable_set_min_size(cp_hashtable *table, int min_size)
{
    int syncbit;
    
    syncbit = table->mode & COLLECTION_MODE_NOSYNC;
    if (!syncbit) cp_hashtable_wrlock(table);
    table->min_size = min_size;
    if (!syncbit) cp_hashtable_unlock(table);
}


void cp_hashtable_set_max_fill_factor(cp_hashtable *table, int fill_factor)
{
    int syncbit;
    
    syncbit = table->mode & COLLECTION_MODE_NOSYNC;
    if (!syncbit) cp_hashtable_wrlock(table);
    table->fill_factor_max = fill_factor;
    if (!syncbit) cp_hashtable_unlock(table);
}

void cp_hashtable_set_min_fill_factor(cp_hashtable *table, int fill_factor)
{
    int syncbit;
    
    syncbit = table->mode & COLLECTION_MODE_NOSYNC;
    if (!syncbit) cp_hashtable_wrlock(table);
    table->fill_factor_min = fill_factor;
    if (!syncbit) cp_hashtable_wrlock(table);
}


/**
 * Initiates a resize of the cp_hashtable which is executed in the background.
 *
 * @return NULL
 */
void *cp_hashtable_resize(cp_hashtable *table, long new_size)
{
    cp_hashtable_entry **new_table;

    if (table->resizing) return NULL;

#if 0
    if ((table->items * 100) > (table->table_size * table->fill_factor_max))  /* double table_size if load over (fill_factor_max) percent */
        new_size = cp_hashtable_choose_size(table->table_size * 2);
     else if ((table->items * 100) < (table->table_size * table->fill_factor_min))  /* half table_size if load under (fill_factor_min) percent */
        new_size = cp_hashtable_choose_size(table->table_size / 2);
#endif

    new_table = (cp_hashtable_entry **) malloc(new_size * sizeof(cp_hashtable_entry *));
    if (new_table == NULL) 
    {
        errno = ENOMEM;
        return NULL;
    }
    memset(new_table, 0, new_size * sizeof(cp_hashtable_entry *));

    table->resize_table = table->table;
    table->resize_len = table->table_size;
    table->table = new_table;
    table->table_size = new_size;
    table->resizing = 1;

    pthread_create(&table->resize_thread, NULL, cp_hashtable_resize_thread, table);
    pthread_detach(table->resize_thread);

    return NULL;
}

/**
 * Resizes the table to a new size.
 * 
 * This is invoked by the insertion code if the load * factor goes over the
 * fill factor limits.
 * @param table the object
 * @param new_size desired size.
 *        The system trys to optmize the actual size for good distribution.
 */
void *cp_hashtable_resize_nosync(cp_hashtable *table, long new_size)
{
    long old_size;
    cp_hashtable_entry **old_table;
    cp_hashtable_entry *entry, *next, **insert;
    long i, index;

    old_size = table->table_size;
    old_table = table->table;
    
    table->table_size = cp_hashtable_choose_size(new_size);

#ifdef __TRACE__
    DEBUGMSG("resizing table (nosync): %d to %d\n", old_size, table->table_size);
#endif

    table->table = (cp_hashtable_entry **) malloc(table->table_size * sizeof(cp_hashtable_entry *));
    memset(table->table, 0, table->table_size * sizeof(cp_hashtable_entry *));

    if (table->table == NULL) 
    {
        errno = ENOMEM;
        return NULL;
    }

    for (i = 0; i < old_size; i++) 
    {
        entry = old_table[i];
        while (entry != NULL) 
        {
            index = entry->cp_hashcode % table->table_size;
            next = entry->next;
            entry->next = NULL;
            insert = &table->table[index];
            while (*insert != NULL) 
                insert = &(*insert)->next;

            *insert = entry;
            
            entry = next;
        }
    }

    free(old_table);

    return table;
}
 
/**
 * Internal replace an existing entry with a new key, value pair.
 *
 * @param table the object
 * @param key    Key to search for.
 * @param value  new value
 * @param code   Hash code of the Key (saves recalculating it)
 * @param option operation mode
 * @param resize 0: search in the normal table, 1: search in the resize table
 * @return pointer to table of entries 
 */
static cp_hashtable_entry **cp_hashtable_replace_internal(cp_hashtable *table, 
                                                          void *key, 
                                                          void *value, 
                                                          long code, 
                                                          int option,
                                                          int resize)
{
    cp_hashtable_entry **entry;
    long index;

    index = resize ? code % table->resize_len : code % table->table_size;
    
    entry = resize ? &table->resize_table[index] : &table->table[index];
    while (*entry != NULL) 
    {
#if CP_HASHTABLE_MULTIPLE_VALUES
        if (!(option & COLLECTION_MODE_MULTIPLE_VALUES) && 
            (*table->compare_fn)(key, (*entry)->key) == 0) 
#else
        if ((*table->compare_fn)(key, (*entry)->key) == 0) 
#endif
        {
            if (option & COLLECTION_MODE_DEEP) 
            {
                if (table->free_key)
                    (*table->free_key)((*entry)->key);
                if (table->free_value)
                    (*table->free_value)((*entry)->value);
                (*entry)->key = key;
            }
        
            if (option & COLLECTION_MODE_COPY) 
            {
                (*entry)->key = table->copy_key ? (*table->copy_key)(key) : key;
                (*entry)->value = table->copy_value ? (*table->copy_value)(value) : value;
            }
            else
                (*entry)->value = value;

            (*entry)->cp_hashcode = code;
            break;
        }

        entry = &(*entry)->next;
    }

    return entry;
}

inline void *cp_hashtable_put(cp_hashtable *table, void *key, void *value)
{
    return cp_hashtable_put_by_option(table, key, value, table->mode);
}

inline void *cp_hashtable_put_safe(cp_hashtable *table, void *key, void *value)
{
    return cp_hashtable_put_by_option(table, key, value, table->mode | COLLECTION_MODE_DEEP);
}

inline void *cp_hashtable_put_copy(cp_hashtable *table, void *key, void *value)
{
        return cp_hashtable_put_by_option(table, key, value, table->mode | COLLECTION_MODE_COPY);
}


/* actual insertion code */
void *cp_hashtable_put_by_option(cp_hashtable *table, void *key, void *value, int option)
{
    long code;
    cp_hashtable_entry **entry; /* defined ** for when initializing a bucket */ 
    int syncbit;
    void *ret = NULL;

    syncbit = table->mode & COLLECTION_MODE_NOSYNC;
    if (!syncbit) cp_hashtable_wrlock(table);

    if ((option & COLLECTION_MODE_NORESIZE) == 0)
    {
        if ((table->items * 100) > 
            (table->table_size * table->fill_factor_max))
        {
            int new_size = cp_hashtable_choose_size(table->table_size * 2);
            if (syncbit)
                cp_hashtable_resize_nosync(table, new_size);
            else
                cp_hashtable_resize(table, new_size);
        }
    }
    
    code = abs((*table->hash_fn)(key));

    entry = cp_hashtable_replace_internal(table, key, value, code, option, 0);

    if (*entry == NULL && table->resizing) 
    {
        cp_hashtable_entry **resize_entry;

        resize_entry = cp_hashtable_replace_internal(table, key, value, code, option, 1);
        if (*resize_entry != NULL)  /* prevent write */
            entry = resize_entry;

    }
        
    if (*entry == NULL) 
    {
        /* no entry found, entry points at lookup location */
        *entry = cp_hashtable_create_entry(table, option, key, value, code);
        if (*entry) 
            table->items++;
    }
    
    if (*entry)
        ret = (*entry)->value;

    if (!syncbit) cp_hashtable_unlock(table);

    return ret;
}

/**
 * Internal remove an entry from the table by key.
 *
 * Get the value by key and destroy the entry.
 *
 * @param table the object
 * @param key    Key to search for.
 * @param code   Hash code of the Key (saves recalculating it)
 * @param mode   operation mode
 * @param resize 0: search in the normal table, 1: search in the resize table
 * @retval value of the entry with key.
 * @retval NULL otherwise (no entry with given key)
 */
void *cp_hashtable_remove_internal(cp_hashtable *table, void *key, long code, int mode, int resize)
{
    cp_hashtable_entry **entry, *item;
    void *value = NULL;

    entry = resize ? &table->resize_table[code % table->resize_len]
                   : &table->table[code % table->table_size];

    while (*entry != NULL) 
    {
        if ((*table->compare_fn)(key, (*entry)->key) == 0) 
        {
            /* entry now points either to the table->table element or to the  */
            /* next pointer in the parent entry                               */
            table->items--;
            item = *entry;
            *entry = item->next;

            value = item->value;
            if (mode & COLLECTION_MODE_DEEP) 
            {
                if (table->free_key)
                    (*table->free_key)(item->key);
                if (table->free_value)
                    (*table->free_value)(item->value);
            }

            free(item);

#if CP_HASHTABLE_MULTIPLE_VALUES
            if (!(mode & COLLECTION_MODE_MULTIPLE_VALUES)) 
#endif
            break; 
        }

        entry = &(*entry)->next;
    }

    return value;
}

/**
 * Remove an entry from the table by key with locking mode.
 *
 * Get the value by key and destroy the entry.
 *
 * @param table the object
 * @param key    Key to search for.
 * @param mode   operation/locking mode
 * @retval value of the entry with key.
 * @retval NULL otherwise (no entry with given key)
 */
void *cp_hashtable_remove_by_mode(cp_hashtable *table, void *key, int mode)
{
    long code;
    void *value = NULL;
    int syncbit = table->mode & COLLECTION_MODE_NOSYNC;
    if (!syncbit) cp_hashtable_wrlock(table);

    code = abs((*table->hash_fn)(key));

    value = cp_hashtable_remove_internal(table, key, code, mode, 0);

    if (table->resizing) 
    {
        void *resize_value;
        resize_value = cp_hashtable_remove_internal(table, key, code, mode, 1);
        if (value == NULL) value = resize_value;
    }

    if ((table->mode & COLLECTION_MODE_NORESIZE) == 0)
    {
        if (table->table_size > table->min_size &&
            ((table->items * 100) < 
            (table->table_size * table->fill_factor_min)))
        {
            int new_size = cp_hashtable_choose_size(table->table_size / 2);
            if (syncbit)
                cp_hashtable_resize_nosync(table, new_size);
            else
                cp_hashtable_resize(table, new_size);
        }
    }

    if (!syncbit) cp_hashtable_unlock(table);

    return value;
}

void cp_hashtable_remove_all(cp_hashtable *table)
{
    long i;
    cp_hashtable_entry *entry, *next;
    int syncbit = table->mode & COLLECTION_MODE_NOSYNC;
    int deepbit = table->mode & COLLECTION_MODE_DEEP;
    cp_destructor_fn dk = table->free_key;
    cp_destructor_fn dv = table->free_value;

    if (!syncbit) cp_hashtable_wrlock(table);
    
    for (i = 0; i < table->table_size; i++) 
    {
        if ((entry = table->table[i]) != NULL) 
        {
            while (entry != NULL) 
            {
                next = entry->next;
                if (deepbit)
                {
                    if (dk) (*dk)(entry->key);
                    if (dv) (*dv)(entry->value);
                }
                free(entry);
                entry = next;
            }
        }
    }

    if (table->resizing) 
    {
        for (i = 0; i < table->resize_len; i++) 
            if ((entry = table->resize_table[i]) != NULL) 
                while (entry != NULL) 
                {
                    next = entry->next;
                    if (deepbit)
                    {
                        if (dk) (*dk)(entry->key);
                        if (dv) (*dv)(entry->value);
                    }
                    free(entry);
                    entry = next;
                }
    }

    if (!syncbit) cp_hashtable_unlock(table);
}

/* remove using current mode settings */
void *cp_hashtable_remove(cp_hashtable *table, void *key)
{
    return cp_hashtable_remove_by_mode(table, key, table->mode);
}


/**
 * Remove with COLLECTION_MODE_DEEP set
 * @note here cp_hashtable_remove_by_mode returns an invalid pointer on 
 * success, since the item has just been released. We can not use it but we 
 * can tell it isn't null.
 */
int cp_hashtable_remove_deep(cp_hashtable *table, void *key)
{
    return cp_hashtable_remove_by_mode(table, key
              , table->mode | COLLECTION_MODE_DEEP) != NULL;
}

//~~ will leak cp_lists if MULTIPLE_VALUES
int cp_hashtable_contains(cp_hashtable *table, void *key)
{
    int rc = 0;

    if (table != NULL && key != NULL)
    {
        long code;
        void *val;
        int option;
        int syncbit = table->mode & COLLECTION_MODE_NOSYNC; 
        if (!syncbit) cp_hashtable_rdlock(table);

        option = table->mode & COLLECTION_MODE_MULTIPLE_VALUES;
        code = abs((*table->hash_fn)(key));
        val = lookup_internal(table, key, code, 
                              table->mode ^ COLLECTION_MODE_MULTIPLE_VALUES, 0);
        /* when resizing, search also there */
        if (val == NULL && table->resizing) 
            val = lookup_internal(table, key, code, 
                                  table->mode ^ COLLECTION_MODE_MULTIPLE_VALUES, 1);

        rc = val != NULL;

        if (!syncbit) cp_hashtable_unlock(table);
    }
    else
        errno = EINVAL; //~~ and set rc = -1?

    return rc;
}


inline void *cp_hashtable_get(cp_hashtable *table, void *key)
{
    return cp_hashtable_get_by_option(table, key, table->mode);
}

void **cp_hashtable_get_keys(cp_hashtable *table)
{
    long i, j;
    void **keys;
    cp_hashtable_entry *entry = NULL;
    int syncbit; 
    int rc = 0;

    if (table == NULL)
    {
        errno = EINVAL;
        return NULL;
    }

    syncbit = table->mode & COLLECTION_MODE_NOSYNC; 
    if (!syncbit) cp_hashtable_rdlock(table);

    keys = (void **) calloc(table->items, sizeof(void *));
    if (keys == NULL) 
    {
        rc = ENOMEM;
        goto DONE;
    }
    
    for (i = 0, j = 0; i < table->table_size; i++) \
    {
        entry = table->table[i];
        while (entry != NULL) 
        {
            keys[j++] = entry->key;
            entry = entry->next;
        }
    }

    if (table->resizing) 
        for (i = 0; i < table->resize_len; i++) 
            entry = table->resize_table[i];
            while (entry != NULL) 
            {
                keys[j++] = entry->key;
                entry = entry->next;
            }
            
DONE:
    if (!syncbit) cp_hashtable_unlock(table);
    if (rc) errno = rc;

    return keys;
}
    

long cp_hashtable_count(cp_hashtable *table)
{
    return (table) ? table->items : 0L; 
}


void **cp_hashtable_get_values(cp_hashtable *table)
{
    long i, j;
    void **values;
    cp_hashtable_entry *entry;
    int syncbit = table->mode & COLLECTION_MODE_NOSYNC;
    int rc = 0;

    if (!syncbit) cp_hashtable_rdlock(table);

    values = (void **) malloc(table->items * sizeof(void *));
    if (values == NULL) 
    {
        rc = ENOMEM;
        goto DONE;
    }
    
    for (i = 0, j = 0; i < table->table_size; i++) 
    {
        entry = table->table[i];
        while (entry != NULL) 
        {
            values[j++] = entry->value;
            entry = entry->next;
        }
    }        

    if (table->resizing) 
    {
        for (i = 0; i < table->resize_len; i++) 
        {
            entry = table->resize_table[i];
            while (entry != NULL) 
            {
                values[j++] = entry->value;
                entry = entry->next;
            }
        }
    }

DONE:
    if (!syncbit) cp_hashtable_unlock(table);
    if (rc) errno = rc;

    return values;
}


/* ------------------------------------------------------------------------ */
/* ---   hash function implementations for primitive types & cp_strings    --- */
/* ------------------------------------------------------------------------ */

long cp_hash_int(void *i)
{
    return (long) *((int *) i);
}


int cp_hash_compare_int(void *i, void *j)
{
    return *((int *) i) - *((int *) j);
}

long cp_hash_long(void *l)
{
    return *((long *) l);
}


int cp_hash_compare_long(void *i, void *j)
{
    long diff = *((long *) i) - *((long *) j);
    return diff > 0 ? 1 : diff < 0 ? -1 : 0;
}

long cp_hash_string(void *str)
{
    int i;
    long res;
    char *_str;
    
    if (str == NULL) return 0; 

    _str = (char *) str;
    
    for (i = 0, res = 1; *_str != '\0'; _str++)
        res = res * HASH_SEED + *_str;

    return res;
}
    

int cp_hash_compare_string(void *s1, void *s2)
{
    if (s1 == NULL && s2 == NULL) return 0;
    if (s1 == NULL || s2 == NULL) return -1;
    return strcmp((char *) s1, (char *) s2);
}

#define CP_CHAR_UC(x) ((x) >= 'a' && (x) <= 'z' ? ((x) - 'a' + 'A') : (x))
        
long cp_hash_istring(void *str)
{
    int i;
    long res;
    char *_str;
    
    if (str == NULL) return 0; 

    _str = (char *) str;
    
    for (i = 0, res = 1; *_str != '\0'; _str++)
        res = res * HASH_SEED + CP_CHAR_UC(*_str);

    return res;
}

int cp_hash_compare_istring(void *s1, void *s2)
{
    if (s1 == NULL && s2 == NULL) return 0;
    if (s1 == NULL || s2 == NULL) return -1;
    return strcasecmp((char *) s1, (char *) s2);
}

void *cp_hash_copy_string(void *element)
{
    char *src;
    char *dst;
    
    src = (char *) element;
    dst = (char *) malloc((strlen(src) + 1) * sizeof(char));

    if (dst == NULL) return NULL;

    strcpy(dst, src);
    return dst;
}

/** @} */

