/* supress warning for strndup */
#if (defined linux || defined __linux || defined __gnu_linux)
#define _GNU_SOURCE
#endif

#include <string.h>
#include <strings.h>
#include <regex.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef USE_SSL
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif

/**
 * @addtogroup cp_socket
 */
/** @{ */
/**
 * @file
 * http framework implementation
 */


#include "common.h"
#include "http.h"
#include "hashtable.h"
#include "hashlist.h"
#include "trie.h"
#include "vector.h"
#include "log.h"
#include "util.h"
#include "socket.h"

/****************************************************************************
 *                                                                          *
 *                    constants and initialization code                     *
 *                                                                          *
 ****************************************************************************/

#ifdef __OpenBSD__
#define HTTP_PARSE_BUFSIZE 0x1000
#else
#define HTTP_PARSE_BUFSIZE 0xFFFF
#endif

#define MATCHMAX 10 

static char *re_request_line_str = 
	"^(OPTIONS|GET|HEAD|POST|PUT|DELETE|TRACE|CONNECT)[[:space:]]+([^ ]+)[[:space:]]+HTTP/([^[:space:]]*)[[:space:]]*$";
static char *re_request_header_str = 
	"^([[:alnum:][:punct:]]+): (.*[^[:space:]])[[:space:]]*$";
static char *re_request_vars_str = "([^&]+)=([^&]*[^&[:space:]])&?";

static regex_t re_request_line;
static regex_t re_request_header;
static regex_t re_request_vars;

static char *cp_http_version_lit[] = { "1.0", "1.1" };

static char *connection_policy_lit[] = { "Close", "Close", "Keep-Alive" };

static char *content_type_lit[] = { "text/plain", "text/html", "image/jpeg" };

static char *server_error_fmt = 
	"<html>\n"
	"<head>\n"
	"<title> 400 BAD REQUEST </title>\n"
	"</head>\n"
	"<body>\n"
	"<h1> 400 BAD REQUEST: %s </h1>\n"
	"</body>\n"
	"</html>";

static int cp_http_server_error_code[] =
	{
		CP_HTTP_EMPTY_REQUEST,
		CP_HTTP_INVALID_REQUEST_LINE,
		CP_HTTP_UNKNOWN_REQUEST_TYPE,
		CP_HTTP_INVALID_URI,
		CP_HTTP_VERSION_NOT_SPECIFIED,
		CP_HTTP_1_1_HOST_NOT_SPECIFIED,
		CP_HTTP_INCORRECT_REQUEST_BODY_LENGTH,
		-1
	};
	
static char *cp_http_server_error_lit[] = 
	{
		"EMPTY REQUEST",
		"INVALID REQUEST LINE",
		"UNKNOWN REQUEST TYPE",
		"INVALID URI",
		"VERSION NOT SPECIFIED",
		"HTTP/1.1 HOST NOT SPECIFIED",
		"INCORRECT REQUEST BODY LENGTH"
	};

static cp_hashtable *cp_http_server_error;

struct cp_http_status_code_entry 
{
	cp_http_status_code code;
	char *lit;
} cp_http_status_code_list[] = 
	{
		{ HTTP_100_CONTINUE, "100 CONTINUE" },
		{ HTTP_101_SWITCHING_PROTOCOLS, "101 SWITCHING PROTOCOLS" },
		{ HTTP_200_OK, "200 OK" },
		{ HTTP_201_CREATED, "201 CREATED" },
		{ HTTP_202_ACCEPTED, "202 ACCEPTED" },
		{ HTTP_203_NON_AUTHORITATIVE_INFORMATION, "203 NON AUTHORITATIVE INFORMATION" },
		{ HTTP_204_NO_CONTENT, "204 NO CONTENT" },
		{ HTTP_205_RESET_CONTENT, "205 RESET CONTENT" },
		{ HTTP_206_PARTIAL_CONTENT, "206 PARTIAL CONTENT" },
		{ HTTP_300_MULTIPLE_CHOICES, "300 MULTIPLE CHOICES" },
		{ HTTP_301_MOVED_PERMANENTLY, "301 MOVED PERMANENTLY" },
		{ HTTP_302_FOUND, "302 FOUND" },
		{ HTTP_303_SEE_OTHER, "303 SEE OTHER" },
		{ HTTP_304_NOT_MODIFIED, "304 NOT MODIFIED" },
		{ HTTP_305_USE_PROXY, "305 USE PROXY" },
		{ HTTP_307_TEMPORARY_REDIRECT, "307 TEMPORARY REDIRECT" },
		{ HTTP_400_BAD_REQUEST, "400 BAD REQUEST" },
		{ HTTP_401_UNAUTHORIZED, "401 UNAUTHORIZED" },
		{ HTTP_402_PAYMENT_REQUIRED, "402 PAYMENT REQUIRED" },
		{ HTTP_403_FORBIDDEN, "403 FORBIDDEN" },
		{ HTTP_404_NOT_FOUND, "404 NOT FOUND" },
		{ HTTP_405_METHOD_NOT_ALLOWED, "405 METHOD NOT ALLOWED" },
		{ HTTP_406_NOT_ACCEPTABLE, "406 NOT ACCEPTABLE" },
		{ HTTP_407_PROXY_AUTHENTICATION_REQUIRED, "407 PROXY AUTHENTICATION REQUIRED" },
		{ HTTP_408_REQUEST_TIME_OUT, "408 REQUEST TIME OUT" },
		{ HTTP_409_CONFLICT, "409 CONFLICT" },
		{ HTTP_410_GONE, "410 GONE" },
		{ HTTP_411_LENGTH_REQUIRED, "411 LENGTH REQUIRED" },
		{ HTTP_412_PRECONDITION_FAILED, "412 PRECONDITION FAILED" },
		{ HTTP_413_REQUEST_ENTITY_TOO_LARGE, "413 REQUEST ENTITY TOO LARGE" },
		{ HTTP_414_REQUEST_URI_TOO_LARGE, "414 REQUEST URI TOO LARGE" },
		{ HTTP_415_UNSUPPORTED_MEDIA_TYPE, "415 UNSUPPORTED MEDIA TYPE" },
		{ HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE, "416 REQUESTED RANGE NOT SATISFIABLE" },
		{ HTTP_417_EXPECTATION_FAILED, "417 EXPECTATION FAILED" },
		{ HTTP_500_INTERNAL_SERVER_ERROR, "500 INTERNAL SERVER ERROR" },
		{ HTTP_501_NOT_IMPLEMENTED, "501 NOT IMPLEMENTED" },
		{ HTTP_502_BAD_GATEWAY, "502 BAD GATEWAY" },
		{ HTTP_503_SERVICE_UNAVAILABLE, "503 SERVICE UNAVAILABLE" },
		{ HTTP_504_GATEWAY_TIME_OUT, "504 GATEWAY TIME OUT" },
		{ HTTP_505_HTTP_VERSION_NOT_SUPPORTED, "505 HTTP VERSION NOT SUPPORTED" },
		{ -1, NULL}
	};

static cp_hashtable *cp_http_status_lit;

static char *cp_http_request_type_lit[] = 
	{
		"OPTIONS", 
		"GET", 
		"HEAD", 
		"POST", 
		"PUT", 
		"DELETE", 
		"TRACE", 
		"CONNECT" 
	};

static volatile int initialized = 0;

static volatile int socket_reg_id = 0;
static cp_hashlist *socket_registry = NULL;

static int log_regex_compilation_error(int rc, char *msg)
{
	char *smsg = regex_compilation_error(rc);
	cp_error(CP_REGEX_COMPILATION_FAILURE, "%s: %s", msg, smsg);
	return rc;
}

void cp_http_signal_handler(int sig);
			
void *session_cleanup_thread(void *);

int cp_http_init()
{
	int i;
	int rc;
	struct sigaction act;

	if (initialized) return initialized++;
	initialized = 1;

	cp_socket_init();

	act.sa_handler = cp_http_signal_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGPIPE, &act, NULL);
	sigaction(SIGTERM, &act, NULL);

	if ((rc = regcomp(&re_request_line, re_request_line_str, 
			REG_EXTENDED | REG_NEWLINE)))
		return log_regex_compilation_error(rc, "cp_http_init - request line");
	
	if ((rc = regcomp(&re_request_header, re_request_header_str, 
			REG_EXTENDED | REG_NEWLINE)))
		return log_regex_compilation_error(rc, "cp_http_init - request header");

	if ((rc = regcomp(&re_request_vars, re_request_vars_str, 
			REG_EXTENDED | REG_NEWLINE)))
		return log_regex_compilation_error(rc, "cp_http_init - request variables");

	cp_http_server_error = cp_hashtable_create(10, cp_hash_int, cp_hash_compare_int);
	for (i = 0; cp_http_server_error_code[i] != -1; i++)
		cp_hashtable_put(cp_http_server_error, 
					  &cp_http_server_error_code[i], 
					  cp_http_server_error_lit[i]);

	cp_http_status_lit = cp_hashtable_create(50, cp_hash_int, cp_hash_compare_int);
	for (i = 0; cp_http_status_code_list[i].lit != NULL; i++)
		cp_hashtable_put(cp_http_status_lit,
					  &cp_http_status_code_list[i].code, 
				      cp_http_status_code_list[i].lit);
	
	socket_registry = 
		cp_hashlist_create_by_mode(1, COLLECTION_MODE_DEEP, 
				                   cp_hash_int, cp_hash_compare_int);

#ifdef USE_HTTP_SESSIONS
	{
		pthread_t _session_cleanup_thread;
		pthread_create(&_session_cleanup_thread, NULL, 
				       session_cleanup_thread, NULL);
		pthread_detach(_session_cleanup_thread);
	}
#endif
	
	return 0;
}

/* signal all sockets in select() to stop */
void cp_httpsocket_stop_all()
{
	cp_socket_stop_all();
}

static volatile int cp_http_shutting_down = 0;

void cp_httpsocket_stop(cp_httpsocket *sock)
{
	if (sock) sock->sock->closing = 1;
}

void cp_http_shutdown()
{
	if (cp_http_shutting_down) return; /* one shutdown should be enough */

	cp_http_shutting_down = 1;

	DEBUGMSG("shutting down http services");
	if (initialized)
	{
		regfree(&re_request_line);
		regfree(&re_request_header);
		regfree(&re_request_vars);
	}

	cp_hashtable_destroy(cp_http_server_error);
	cp_hashtable_destroy(cp_http_status_lit);
	cp_hashlist_destroy_custom(socket_registry, NULL, 
							   (cp_destructor_fn) cp_httpsocket_stop);
	socket_registry = NULL;

	cp_socket_shutdown();

	initialized = 0;
}

void cp_http_signal_handler(int sig)
{
	switch (sig)
	{
		case SIGINT: 
			DEBUGMSG("SIGINT - stopping\n");
			cp_http_shutdown();
			break;

		case SIGPIPE: 
			DEBUGMSG("SIGPIPE - ignoring\n");
			break;

		case SIGTERM:
			DEBUGMSG("SIGTERM - stopping\n");
			cp_http_shutdown();
			break;
	}
}

#define HTTP_RE_MATCHMAX 10

#ifdef USE_HTTP_SESSIONS

/****************************************************************************
 *                                                                          *
 *  cp_http_session_* functions:                                            *
 *                                                                          *
 *  o cp_httpsocket_create_session - creates a session [static]             *
 *  o cp_httpsocket_checkout_session - retrieves/creates session [static]   *
 *  o cp_httpsocket_remove_session - drops a session [static]               *
 *  o cp_http_session_delete - deletes a session [static]                   *
 *  o cp_http_session_inc_refcount - increase reference count               *
 *  o cp_http_session_dec_refcount - decrease reference count               *
 *  o cp_http_session_get - retrieves a value by key                        *
 *  o cp_http_session_set - sets a value that doesn't need to be deleted    *
 *  o cp_http_session_set_dtr - sets a value with a destructor callback     *
 *  o cp_http_session_set_validity - sets session validity time             *
 *  o cp_http_session_is_fresh - session is newly created                   *
 *                                                                          *
 *  client code should use cp_http_request_get_session to get a reference   *
 *  to the session. use cp_http_session_increase_refcount and               *
 *  cp_http_session_decrease_refcount only if performing special tricks.    *
 *                                                                          *
 ****************************************************************************/

static void cp_http_session_delete(cp_http_session *session);

static cp_http_session *cp_httpsocket_create_session(cp_httpsocket *sock)
{
	cp_http_session *session = calloc(1, sizeof(cp_http_session));
	if (session)
	{
		struct timeval tm;
		gettimeofday(&tm, NULL);
		session->sid = malloc(17 * sizeof(char));
		if (session->sid == NULL)
		{
			free(session);
			return NULL;
		}
		session->key = 
			cp_hashtable_create_by_option(COLLECTION_MODE_COPY |
                                          COLLECTION_MODE_DEEP,
										  1,
										  cp_hash_string, 
										  cp_hash_compare_string,
										  (cp_copy_fn) strdup, 
										  (cp_destructor_fn) free,
										  NULL, 
										  (cp_destructor_fn) cp_wrap_delete);
		if (session->key == NULL)
		{
			free(session->sid);
			free(session);
			return NULL;
		}

		session->created = session->access = tm.tv_sec;
		session->validity = DEFAULT_SESSION_VALIDITY;
		session->fresh = 1;
		session->refcount = 1;
		session->valid = 1;

		if (sock->session == NULL)
		{
			sock->session = 
				cp_hashlist_create_by_option(COLLECTION_MODE_COPY |
					                         COLLECTION_MODE_DEEP,
									         1,
										     cp_hash_string,
										     cp_hash_compare_string, 
										     (cp_copy_fn) strdup,
										     (cp_destructor_fn) free,
										     NULL,
										     (cp_destructor_fn) 
											     cp_http_session_delete);
			if (sock->session == NULL)
			{
				cp_http_session_delete(session);
				return NULL;
			}
		}
		//~~ although improbable (~ 1:100,000), gen_tm_id_str may return 
		// doubles and should be replaced with a more robust function. until
		// then, make sure the id isn't already in there.
		do 
		{ 
			gen_tm_id_str(session->sid, &tm);
		} while (cp_hashlist_get(sock->session, session->sid));

		cp_hashlist_append(sock->session, session->sid, session);
	}

	return session;
}

static void cp_httpsocket_remove_session(cp_httpsocket *sock, char *sid)
{
	cp_http_session *session = 
		sock->session ? cp_hashlist_get(sock->session, sid) : NULL;

	if (session)
	{
		if (--session->refcount <= 0)
		{
			cp_hashlist_remove(sock->session, sid); /* also calls ~session() */
			if (sock->pending)
			{
				cp_http_response *res = 
					cp_hashlist_remove(sock->pending, sid);
				if (res) 
					cp_http_response_delete(res);
			}
		}
	}
}
		
static cp_http_session *
	cp_httpsocket_checkout_session(cp_httpsocket *sock, char *sid)
{
	cp_http_session *session = 
		sock->session ? cp_hashlist_get(sock->session, sid) : NULL;

	if (session)
		++session->refcount;
    //~~ else create invalid session with specified id?

	return session;
}

/**
 * internal ~session() function, called by cp_httpsocket_remove_session when
 * refcount hits 0 - do not invoke directly. 
 */
static void cp_http_session_delete(cp_http_session *session)
{
	if (session)
	{
		if (session->sid) free(session->sid);
		cp_hashtable_destroy(session->key); /* calls ~cp_wrap() */
		free(session);
	}
}

void *cp_http_session_get(cp_http_session *session, char *key)
{
	cp_wrap *wrap = cp_hashtable_get(session->key, key);
	if (wrap)
		return wrap->item;
	return NULL;
}

void *cp_http_session_set_dtr(cp_http_session *session, 
		                      char *key, void *value, 
							  cp_destructor_fn dtr)
{
	cp_wrap *wrap = cp_wrap_new(value, dtr);
	if (wrap)
		return cp_hashtable_put(session->key, key, wrap);
	return NULL;
}

inline void *cp_http_session_set(cp_http_session *session, 
		                         char *key, void *value)
{
	return cp_http_session_set_dtr(session, key, value, NULL);
}

inline void cp_http_session_set_validity(cp_http_session *session, long sec)
{
	session->validity = sec;
}

inline short cp_http_session_is_fresh(cp_http_session *session)
{
	return session->fresh;
}

static char *redirect_fmt = "<html><head><title>302 FOUND</title></head><body><h1>302 FOUND</h1>The document has moved <a href=\"%s\"> here </a>.</body></html>";

/**
 * send a redirect and set a session id cookie. This is captured on the next
 * client request, and the original request is then written to the client.
 */
static cp_http_response *
	cp_http_new_session_redirect(cp_http_request *req, int step)
{
	cp_http_response *res = cp_http_response_create(req);
	
	if (res)
	{
		char redirect_url[MAX_URL_LENGTH]; 
#ifdef USE_COOKIES
		char cookie_path[MAX_URL_LENGTH]; 
		char *cp;
#endif /* USE_COOKIES */
		char *host; 
		char buf[HTTP_PARSE_BUFSIZE];
		cp_string *content;
		char *protocol;

#ifdef USE_SSL
		if (req->owner->sock->use_ssl)
			protocol = "https";
		else
#endif
		protocol = "http";

		if ((host = cp_http_request_get_header(req, "Host")) == NULL)
			host = "localhost"; //~~ figure out hostname
		
#ifdef USE_COOKIES
		if (step == 2)
			sprintf(redirect_url, "%s://%s%s", protocol, host, req->uri);
		else
#endif /* USE_COOKIES */
		sprintf(redirect_url, "%s://%s%s?%s=%s", protocol, host, req->uri,
				CP_HTTP_SESSION_PRM, req->session->sid);

		cp_http_response_set_status(res, HTTP_302_FOUND);
		cp_http_response_set_header(res, "Location", redirect_url);
		cp_http_response_set_connection_policy(res, 
				HTTP_CONNECTION_POLICY_KEEP_ALIVE);

		sprintf(buf, redirect_fmt, redirect_url);
		content = cp_string_cstrcpy(buf);
		cp_http_response_set_content(res, content);

#ifdef USE_COOKIES
		if (step == 1)
		{
			strcpy(cookie_path, req->uri);
			if ((cp = strrchr(cookie_path, '/')))
			{
				cp++;
				*cp = '\0';
			}
		
			cp_http_response_set_cookie(res, CP_HTTP_SESSION_PRM, 
					req->session->sid, host, cookie_path, 
					req->session->validity, 
#ifdef USE_SSL
				req->owner->sock->use_ssl
#else
				0
#endif /* USE_SSL */
				);
		}
#endif /* USE_COOKIES */
	}

	return res;
}

/** session cleanup thread function */
void *session_cleanup_thread(void *prm)
{
	cp_hashlist_iterator i, j;
	cp_httpsocket *curr;
	cp_http_session *session;
	struct timeval delay;
	time_t now;
	int rc;

	while (!cp_http_shutting_down)
	{
		delay.tv_sec = 900; /* run session cleanup every 15 minutes */
		delay.tv_usec = 0;
		rc = select(0, NULL, NULL, NULL, &delay);  //~~ pthread_cond_timedwait?
		if (cp_http_shutting_down) break;
#ifdef __TRACE__
		DEBUGMSG("session_cleanup_thread running");
#endif
		now = time(NULL);
		cp_hashlist_iterator_init(&i, socket_registry, COLLECTION_LOCK_READ);
		while ((curr = cp_hashlist_iterator_next_value(&i)))
		{
			if (curr->session == NULL) continue;
			cp_hashlist_iterator_init(&j, curr->session, COLLECTION_LOCK_WRITE);
			while ((session = cp_hashlist_iterator_next_value(&j)) != NULL)
			{
				if (now > session->access + session->validity) /* expired */
					cp_httpsocket_remove_session(curr, session->sid);
			}
			cp_hashlist_iterator_release(&j); /* release locks */
		}
		cp_hashlist_iterator_release(&i); /* release locks here too */
	}
#ifdef __TRACE__
	DEBUGMSG("session_cleanup_thread exits");
#endif
		
	return NULL;
}
#endif /* USE_HTTP_SESSIONS */

/****************************************************************************
 *                                                                          *
 *  cp_http_request_* functions:                                            *
 *                                                                          *
 *  o cp_http_request_delete - performs cleanup                             *
 *  o cp_http_request_parse - parses a request from a char buffer           *
 *  o cp_http_request_get_header - returns headers by name                  *
 *  o cp_http_request_dump - dumps request to stdout and log                *
 *                                                                          *
 ****************************************************************************/

void cp_http_request_delete(cp_http_request *request)
{
	if (request)
	{
		free(request->uri);
#ifdef USE_COOKIES
		if (request->cookie)
			cp_vector_destroy_custom(request->cookie, free);
#endif /* USE_COOKIES */
		if (request->header)
			cp_hashtable_destroy(request->header);
		if (request->prm)
			cp_hashtable_destroy(request->prm);
		free(request);
	}
}

static int parse_request_line(cp_http_request *req, char *request, char **cgivars_lit, int *plen)
{
	regmatch_t rm[HTTP_RE_MATCHMAX];
	int len;
	int valid;

	if (regexec(&re_request_line, request, HTTP_RE_MATCHMAX, rm, 0))
		return CP_HTTP_INVALID_REQUEST_LINE;

	valid = 0;
	len = rm[1].rm_eo - rm[1].rm_so;
	if (len)
		for (req->type = OPTIONS; req->type <= CONNECT; req->type++)
			if (strncasecmp(&request[rm[1].rm_so], 
				cp_http_request_type_lit[req->type], len) == 0)
			{
				valid = 1;
				break;
			}

	if (!valid) /* unknown or missing method */
		return CP_HTTP_UNKNOWN_REQUEST_TYPE;

	valid = 0;
	len = rm[2].rm_eo - rm[2].rm_so;
	if (len)
	{
		char *q = strnchr(&request[rm[2].rm_so], '?', rm[0].rm_eo);
		if (q) 
		{
			//~~ check if this is a GET - if not, reject
			len = q - request;
			*cgivars_lit = strndup(q + 1, rm[2].rm_eo - len);
			len -= rm[2].rm_so;
		}

		req->uri = strndup(&request[rm[2].rm_so], len);
		valid = 1;
	}
	
	if (!valid) /* invalid uri */
		return CP_HTTP_INVALID_URI;
	
	valid = 0;
	len = rm[3].rm_eo - rm[3].rm_so;

	if (len)
	{
		if (strncmp(&request[rm[3].rm_so], "1.0", len) == 0)
		{
			req->version = HTTP_1_0;
			valid = 1;
		}
		else if (strncmp(&request[rm[3].rm_so], "1.1", len) == 0)
		{
			req->version = HTTP_1_1;
			valid = 1;
		}
	}

	if (!valid) /* HTTP version not specified */
		return CP_HTTP_VERSION_NOT_SPECIFIED;

	*plen = rm[0].rm_eo;

	return 0;
}

static int parse_headers(cp_http_request *req, char *request, int *plen)
{
	char *curr = request;
	regmatch_t rm[MATCHMAX];
	int len;
	char *name, *value;
			
	req->header = 
		cp_hashtable_create_by_option(COLLECTION_MODE_NOSYNC | 
				                      COLLECTION_MODE_DEEP, 10,
									  cp_hash_string, cp_hash_compare_string, 
									  NULL, free, NULL, free);
			
	while (regexec(&re_request_header, curr, MATCHMAX, rm, 0) == 0)
	{
		len = rm[1].rm_eo - rm[1].rm_so;
		if ((name = (char *) malloc((len + 1) * sizeof(char))) == NULL) 
			return CP_MEMORY_ALLOCATION_FAILURE;
		strncpy(name, curr + rm[1].rm_so, len);
		name[len] = '\0';

		len = rm[2].rm_eo - rm[2].rm_so;
		if ((value = (char *) malloc((len + 1) * sizeof(char))) == NULL) 
		{
			free(name);
			return CP_MEMORY_ALLOCATION_FAILURE;
		}
		strncpy(value, curr + rm[2].rm_so, len);
		value[len] = '\0';

#ifdef USE_COOKIES
		if (strcmp(name, "Cookie") == 0)
		{
			if (req->cookie == NULL)
			{
				req->cookie = cp_vector_create(1, sizeof(char *));
				if (req->cookie == NULL)
					return CP_MEMORY_ALLOCATION_FAILURE;
			}
			cp_vector_add_element(req->cookie, value);
			free(name);
		}
		else
#endif /* USE_COOKIES */
		cp_hashtable_put(req->header, name, value);

		curr = &curr[rm[0].rm_eo];
	}
	*plen = curr - request;

	if ((req->version == HTTP_1_1) && 
		(cp_http_request_get_header(req, "Host") == NULL))
		return CP_HTTP_1_1_HOST_NOT_SPECIFIED;
			
	return 0;
}

static int parse_cp_http_request_body(cp_http_request *req, 
								   char *request, 
								   char **cgivars_lit)
{
	int len, dlen;
	char *clen;
	int step = 4;
	char *p = strstr(request, "\r\n\r\n");

	if (p == NULL) { p = strstr(request, "\n\n"); step = 2; }
	if (p == NULL) { cp_warn("p == NULL"); return CP_HTTP_INCORRECT_REQUEST_BODY_LENGTH; }

	p += step;
	len = strlen(p);
	clen = cp_http_request_get_header(req, "Content-Length");
	if (clen == NULL) { cp_warn("no length header"); return CP_HTTP_INCORRECT_REQUEST_BODY_LENGTH; }
	dlen = atoi(clen);
	if (dlen != len) { cp_warn("%d != %d.", len, dlen); return CP_HTTP_INCORRECT_REQUEST_BODY_LENGTH; }

	*cgivars_lit = strdup(p);
	
	return 0;
}

#define HEXDIGIT(c) ((c) >= '0' && (c) <= '9' ? (c) - '0' : \
		             ((c) >= 'A' && (c) <= 'F' ? (c) - 'A' + 0xA : \
		             ((c) >= 'a' && (c) <= 'f' ? (c) - 'a' + 0xA : -1)))

void urldecode(char *str, char **decoded)
{
	char dbuf[HTTP_PARSE_BUFSIZE];
	char *src, *dst;
	int len;

	for (src = str, dst = dbuf, len = 0; *src && len < HTTP_PARSE_BUFSIZE; src++, dst++, len++)
	{
		switch (*src)
		{
			case '+': 
				*dst = ' '; 
				break; 
				
			case '%': 
				src++;
				if (*src == '%') 
					*dst = '%';
				else
					*dst = (HEXDIGIT(src[0])) << 4 | (HEXDIGIT(src[1]));
					src++;
				break;

			default:
				*dst = *src;
				break;
		}
	}

	*dst = '\0';
	
	*decoded = strndup(dbuf, len);
}

static int parse_cgi_vars(cp_http_request *req, char *cgivars_lit)
{
	regmatch_t rm[MATCHMAX];
	int len;
	char *pos;
	char *raw_name, *raw_value;
	char *name, *value;

	req->prm = 
		cp_hashtable_create_by_option(COLLECTION_MODE_NOSYNC | 
				                      COLLECTION_MODE_DEEP, 10,
									  cp_hash_string, cp_hash_compare_string,
									  NULL, (cp_destructor_fn) free,
									  NULL, (cp_destructor_fn) free);

	for (pos = cgivars_lit; 
		 regexec(&re_request_vars, pos, MATCHMAX, rm, 0) == 0; 
		 pos = &pos[rm[0].rm_eo])
	{
		len = rm[1].rm_eo - rm[1].rm_so;
		raw_name = strndup(&pos[rm[1].rm_so], len);

		len = rm[2].rm_eo - rm[2].rm_so;
		raw_value = strndup(&pos[rm[2].rm_so], len);

		urldecode(raw_name, &name);
		urldecode(raw_value, &value);
		free(raw_name);
		free(raw_value);
		
		cp_hashtable_put(req->prm, name, value);
	}

	return 0;
}

cp_http_request *cp_http_request_parse(cp_httpsocket *owner, 
		                               char *request, 
									   int *err)
{
	cp_http_request *req = NULL; // prevents gcc warning
	char *cgivars_lit = NULL;
	int rc;
	int len;
	char *curr;
	char *clen;

	if (!initialized) cp_http_init();

	if (request == NULL || *request == '\0') 
	{
		rc = CP_HTTP_EMPTY_REQUEST;
		goto ERROR;
	}

	req = (cp_http_request *) calloc(1, sizeof(cp_http_request));
	if (req == NULL)
	{
		rc = CP_MEMORY_ALLOCATION_FAILURE;
		goto ERROR;
	}
	
	/* parse request line (eg GET / HTTP/1.0) */
	if ((rc = parse_request_line(req, request, &cgivars_lit, &len)))
		goto ERROR;

	/* parse headers */
	curr = &request[len];
	if ((rc = parse_headers(req, curr, &len)))
		goto ERROR;

	/* read request body if any */
	if ((clen = cp_http_request_get_header(req, "Content-Length")))
	{
		curr = &curr[len];
		if ((rc = parse_cp_http_request_body(req, curr, &cgivars_lit)))
			goto ERROR;
	}

	/* parse cgi parameters */
	if (cgivars_lit)
	{
		if ((rc = parse_cgi_vars(req, cgivars_lit)))
			goto ERROR;

		free(cgivars_lit);
	}

	/* session handling if enabled */
#ifdef USE_HTTP_SESSIONS
	{
		char *sid;
#ifdef USE_COOKIES
		if (req->cookie) //~~ move this to cookie parsing to avoid looping?
		{
			char *cookie;
			int i;
			int n = cp_vector_size(req->cookie);
			for (i = 0; i < n; i++)
			{
				cookie = cp_vector_element_at(req->cookie, i);
				if (strncmp(cookie, 
					CP_HTTP_SESSION_MARKER, CP_HTTP_SESSION_MARKER_LEN) == 0)
				{
					sid = &cookie[CP_HTTP_SESSION_MARKER_LEN];
					req->session = cp_httpsocket_checkout_session(owner, sid);
					if (req->session == NULL)
						cp_warn("can\'t find session record %s", sid);
					else
					{
						req->session->type = SESSION_TYPE_COOKIE;
						req->session->access = time(NULL);
//						req->session->fresh = 0;
					}
				}
			}
		}
		if (req->session == NULL)
#endif /* USE_COOKIES */
		{
			sid = cp_http_request_get_parameter(req, CP_HTTP_SESSION_PRM);
			if (sid)
			{
				req->session = cp_httpsocket_checkout_session(owner, sid);
				if (req->session == NULL)
					cp_warn("can\'t find session record %s", sid);
				else
				{
					req->session->type = SESSION_TYPE_URLREWRITE;
					req->session->access = time(NULL);
				}
			}
		}
	}
#endif /* USE_HTTP_SESSIONS */
	
	*err = 0;
	req->owner = owner;
	return req;

ERROR:
	*err = rc;
	if (cgivars_lit) free(cgivars_lit);
	free(req);
	return NULL;
}

char *cp_http_request_get_header(cp_http_request *req, char *header)
{
	char *res = NULL;
	
	if (req && header)
		res = cp_hashtable_get(req->header, header);

	return res;
}

char *cp_http_request_get_parameter(cp_http_request *request, char *name)
{
	char *res = NULL;

	if (request && name && request->prm)
		res = cp_hashtable_get(request->prm, name);

	return res;
}

#ifdef USE_HTTP_SESSIONS
cp_http_session *
	cp_http_request_get_session(cp_http_request *request, int create)
{
	if (request->session == NULL && create)
		request->session = cp_httpsocket_create_session(request->owner);

	return request->session;
}
#endif /* USE_HTTP_SESSIONS */

void cp_http_request_dump(cp_http_request *req)
{
	int i;
	char **keys;
	
	cp_info("dumping http request (%lX)", (long) req);
	cp_info("--------------------------------------------------------------");
	cp_info("method: %s", cp_http_request_type_lit[req->type]);
	cp_info("version: HTTP/%s", cp_http_version_lit[req->version]);
	cp_info("uri: [%s]", req->uri);
	
	cp_info("headers:\n--------");
	keys = (char **) cp_hashtable_get_keys(req->header);
	for (i = 0; i < cp_hashtable_count(req->header); i++)
		cp_info("%s: %s", keys[i], (char *) cp_hashtable_get(req->header, keys[i]));
	free(keys);

#ifdef USE_COOKIES
	if (req->cookie)
	{
		int n;
		cp_info("cookies:\n--------");
		n = cp_vector_size(req->cookie);
		for (i = 0; i < n; i++)
			cp_info("%s", cp_vector_element_at(req->cookie, i));
	}
#endif /* USE_COOKIES */

	if (req->prm)
	{
		cp_info("cgi parameters:\n---------------");
		keys = (char **) cp_hashtable_get_keys(req->prm);
		for(i = 0; i < cp_hashtable_count(req->prm); i++)
			cp_info("%s: %s", keys[i], (char *) cp_hashtable_get(req->prm, keys[i]));
		free(keys);
	}

	cp_info("--------------------------------------------------------------");
}


/****************************************************************************
 *                                                                          *
 *  cp_http_response_* functions:                                           *
 *                                                                          *
 *  o cp_http_response_create - allocates & initialized response struct     *
 *  o cp_http_response_delete - releases a response struct                  *
 *  o cp_http_response_write - writes response to client                    *
 *                                                                          *
 *  o cp_http_response_set_status - set the status code for a response      *
 *  o cp_http_response_set_content_type - set content type code [deprecated]*
 *  o cp_http_response_set_content_type_string - set content type           *
 *  o cp_http_response_set_header - set a header                            *
 *  o cp_http_response_set_body - set the response content [deprecated]     *
 *  o cp_http_response_set_content - set the response content               *
 *  o cp_http_response_set_connection_policy - Close / Keep-Alive           *
 *                                                                          *
 ****************************************************************************/

cp_http_response *cp_http_response_create(cp_http_request *req)
{
	cp_http_response *res = NULL;
	
	if ((res = (cp_http_response *) calloc(1, sizeof(cp_http_response))))
	{
		res->request = req;
        res->header = 
            cp_hashtable_create_by_option(COLLECTION_MODE_NOSYNC |
										  COLLECTION_MODE_COPY |
                                          COLLECTION_MODE_DEEP,
                                          10, 
                                          cp_hash_string, 
                                          cp_hash_compare_string,
                                          (cp_copy_fn) strdup, free, 
                                          (cp_copy_fn) strdup, free);
		if (res->header == NULL)
		{
			free(res);
			return NULL;
		}
		res->status = HTTP_200_OK;
		res->version = req ? req->version : HTTP_1_1;
		res->servername = req->owner->server_name;
	}

	return res;
}

void cp_http_response_delete(cp_http_response *response)
{
	if (response)
	{
		cp_hashtable_destroy(response->header);
#ifdef USE_COOKIES
		if (response->cookie)
			cp_vector_destroy_custom(response->cookie, free);
#endif /* USE_COOKIES */
		if (response->content_type_lit)
			free(response->content_type_lit);
		if (response->content)
			cp_string_delete(response->content);
		if (response->body)
			free(response->body);
		free(response);
	}
}

int cp_http_response_write(cp_connection_descriptor *cdesc, 
		                   cp_http_response *res)
{
	char rbuf[HTTP_PARSE_BUFSIZE];
	char *content = NULL;
	int content_len = 0;
	int len;
	int rc, total;
	int fd = cp_connection_descriptor_get_fd(cdesc);
#ifdef USE_SSL
	cp_socket *sock = cp_connection_descriptor_get_socket(cdesc);
#endif

	if (res->len == 0 && res->body) res->len = strlen(res->body);

	sprintf(rbuf, 
		"HTTP/%s %s\r\n"
		"Server: %s\r\n"
		"Connection: %s\r\n", 
			cp_http_version_lit[res->version], 
			(char *) cp_hashtable_get(cp_http_status_lit, &res->status), 
			res->servername ? res->servername : DEFAULT_SERVER_NAME, 
			connection_policy_lit[res->connection]);

	if (res->header)
	{
		char hbuf[HTTP_PARSE_BUFSIZE];
		int i;
		char **key = (char **) cp_hashtable_get_keys(res->header);
		for (i = 0; i < cp_hashtable_count(res->header); i++)
		{
			sprintf(hbuf, "%s: %s\r\n", 
					key[i], (char *) cp_hashtable_get(res->header, key[i]));
			strcat(rbuf, hbuf);
		}
		free(key);
	}
				
#ifdef USE_COOKIES
	if (res->cookie)
	{
		char cbuf[HTTP_PARSE_BUFSIZE];
		int i;
		for (i = 0; i < cp_vector_size(res->cookie); i++)
		{
			sprintf(cbuf, "Set-Cookie: %s\r\n", 
					(char *) cp_vector_element_at(res->cookie, i));
			strcat(rbuf, cbuf);
		}
	}
#endif /* USE_COOKIES */

	if (res->content)
	{
		char bbuf[HTTP_PARSE_BUFSIZE];
		sprintf(bbuf, "Content-Type: %s\r\nContent-Length: %d\r\n\r\n",
			res->content_type_lit ? 
				res->content_type_lit : content_type_lit[res->content_type],
			res->content->len); //~~
		strcat(rbuf, bbuf);
		content = res->content->data;
		content_len = res->content->len;
	}
	else if (res->body)
	{
		char bbuf[HTTP_PARSE_BUFSIZE];
		content_len = strlen(res->body);
		sprintf(bbuf, "Content-Type: %s\r\nContent-Length: %d\r\n\r\n",
			res->content_type_lit ? 
				res->content_type_lit : content_type_lit[res->content_type],
			content_len);
		strcat(rbuf, bbuf);
		content = res->body;
	}

	len = strlen(rbuf);

#ifdef USE_SSL
	if (sock->use_ssl)
	{
		rc = SSL_write(cdesc->ssl, rbuf, len);
		total = rc;
		if (content_len)
		{
			rc = SSL_write(cdesc->ssl, content, content_len);
			total += rc;
		}
	}
	else
#endif /* USE_SSL */
	{
		rc = write(fd, rbuf, len); //~~ handle errors etc.
		total = rc;
		if (content_len)
		{
			rc = write(fd, content, content_len);
			total += rc;
		}
	}
	
#ifdef __TRACE__
	cp_info("----- http response headers -----\n%s", rbuf);
	if (res->content) 
		cp_ndump(LOG_LEVEL_INFO, res->content, 0x400);
	else if (res->body) 
		cp_nlog(0x400, content);
	if (content_len > 0x400) 
		cp_log("\n ... truncated (%d bytes)\n", content_len);
	cp_info("---------------------------------\n");
#endif

#ifdef DEBUG
	DEBUGMSG("%d bytes written", total);
#endif
	
	return rc;
}

inline void cp_http_response_set_status(cp_http_response *response, 
		                                cp_http_status_code status)
{
	response->status = status;
}

inline void cp_http_response_set_content_type(cp_http_response *response, 
		                                      cp_http_content_type type)
{
	response->content_type = type;
}

inline void cp_http_response_set_content_type_string(cp_http_response *response,
		                                             char *content_type)
{
	response->content_type_lit = content_type ? strdup(content_type) : NULL;
}

inline void cp_http_response_set_header(cp_http_response *response, 
		                                char *name, char *value)
{
	cp_hashtable_put(response->header, name, value);
}

inline void cp_http_response_set_body(cp_http_response *response, 
	                                  char *body)
{
	response->body = strdup(body);
}

inline void cp_http_response_set_content(cp_http_response *response, 
	                                     cp_string *content)
{
	response->content = content;
}

inline void 
	cp_http_response_set_connection_policy(cp_http_response *response, 
		                                   connection_policy policy)
{
	response->connection = policy;
}

#ifdef USE_COOKIES
int cp_http_response_set_cookie(cp_http_response *response, char *name, 
		char *content, char *host, char *path, long validity, int secure)
{
	char cookie[MAX_COOKIE_LENGTH];
	char expiration[0x50];
	time_t now;
	struct tm gmt;
	char *domain;

	now = time(NULL);

	now += validity;
	gmtime_r(&now, &gmt);
	strftime(expiration, 0x50, COOKIE_TIME_FMT, &gmt);

	if (host) 
	{
		if (strncmp(host, "www.", 4) == 0) 
			domain = &host[3];
		else
			domain = NULL; //~~
	}
	else
		domain = NULL;

	sprintf(cookie, "%s=%s; expires=%s", name, content, expiration);
	if (path)
	{
		strcat(cookie, "; path=");
		strcat(cookie, path);
	}
	if (domain)
	{
		strcat(cookie, "; domain=");
		strcat(cookie, domain);
	}
	if (secure) strcat(cookie, "; secure");

#ifdef __TRACE__
	DEBUGMSG("setting cookie (%d chars): %s", strlen(cookie), cookie);
#endif
	if (response->cookie == NULL)
	{
		response->cookie = cp_vector_create(1, sizeof(char *));
		if (response->cookie == NULL) return -1;
	}

	cp_vector_add_element(response->cookie, strdup(cookie)); //~~ vector mode
	
	return 0;
}
#endif

static int cp_http_write_server_error(cp_connection_descriptor *cdesc, 
									  int status_code)
{
	char errbuf[HTTP_PARSE_BUFSIZE];
	char *lit;
	int rc;
	
	cp_http_response *res = cp_http_response_create(NULL);

	if (res == NULL) return CP_MEMORY_ALLOCATION_FAILURE;

	cp_http_response_set_status(res, HTTP_400_BAD_REQUEST);

	lit = cp_hashtable_get(cp_http_server_error, &status_code);

	if (lit)
		sprintf(errbuf, server_error_fmt, lit);
	else
		errbuf[0] = '\0';
	
	cp_http_response_set_content_type(res, HTML);

	res->body = errbuf;

	rc = cp_http_response_write(cdesc, res);

	res->body = NULL;
	cp_http_response_delete(res);
		
	return rc;
}

/****************************************************************************
 *                                                                          *
 * cp_httpsocket_* functions                                                *
 *                                                                          *
 ****************************************************************************/

cp_httpsocket *cp_httpsocket_create(int port, cp_http_service_callback service)
{
	cp_socket *sock;
	cp_httpsocket *psock;

	cp_http_init(); /* idempotent */

	sock = cp_socket_create(port, CPSOCKET_STRATEGY_THREADPOOL, cp_http_thread_fn);
	if (sock == NULL) return NULL; //~~ how about an error code?

	psock = (cp_httpsocket *) calloc(1, sizeof(cp_httpsocket));
	if (psock == NULL)
	{
		cp_socket_delete(sock);
		cp_error(CP_MEMORY_ALLOCATION_FAILURE, "can\'t allocate cp_httpsocket");
	}

	psock->id = socket_reg_id++;
	cp_hashlist_append(socket_registry, &psock->id, psock); /* register */
	
	psock->sock = sock;
	cp_socket_set_owner(sock, psock);

	psock->default_service = service;
	cp_httpsocket_set_keepalive(psock, HTTP_KEEPALIVE);

	return psock;
}

#ifdef USE_SSL
cp_httpsocket *cp_httpsocket_create_ssl(int port, 
		                                cp_http_service_callback service, 
										char *certificate_file,
										char *key_file,
										int verification_mode)
{
	cp_socket *sock;
	cp_httpsocket *psock;

	cp_http_init(); /* idempotent */

	sock = cp_socket_create_ssl(port, CPSOCKET_STRATEGY_THREADPOOL, 
								cp_http_thread_fn, certificate_file, 
								key_file, verification_mode);

	if (sock == NULL) return NULL; //~~ how about an error code?

	psock = (cp_httpsocket *) calloc(1, sizeof(cp_httpsocket));
	if (psock == NULL)
	{
		cp_socket_delete(sock);
		cp_error(CP_MEMORY_ALLOCATION_FAILURE, "can\'t allocate cp_httpsocket");
	}
	
	psock->sock = sock;
	cp_socket_set_owner(sock, psock);

	psock->default_service = service;
	cp_httpsocket_set_keepalive(psock, HTTP_KEEPALIVE);

	return psock;
}
#endif
	
inline void cp_httpsocket_set_keepalive(cp_httpsocket *socket, int sec)
{
	socket->keepalive = sec;
}

inline void cp_httpsocket_set_server_name(cp_httpsocket *socket, char *name)
{
	socket->server_name = strdup(name);
}

inline void cp_httpsocket_set_backlog(cp_httpsocket *socket, int backlog)
{
	cp_socket_set_backlog(socket->sock, backlog);
}

inline void cp_httpsocket_set_delay(cp_httpsocket *socket, struct timeval delay){
	cp_socket_set_delay(socket->sock, delay);
}

inline void cp_httpsocket_set_delay_sec(cp_httpsocket *socket, long sec)
{
	cp_socket_set_delay_sec(socket->sock, sec);
}

inline void cp_httpsocket_set_delay_usec(cp_httpsocket *socket, long usec)
{
	cp_socket_set_delay_usec(socket->sock, usec);
}

inline void cp_httpsocket_set_poolsize_min(cp_httpsocket *socket, int min)
{
	cp_socket_set_poolsize_min(socket->sock, min);
}

inline void cp_httpsocket_set_poolsize_max(cp_httpsocket *socket, int max)
{
	cp_socket_set_poolsize_max(socket->sock, max);
}

void cp_httpsocket_delete(cp_httpsocket *sock)
{
	if (sock)
	{
		if (socket_registry)
			cp_hashlist_remove(socket_registry, &sock->id);

		cp_socket_delete(sock->sock);
#ifdef USE_HTTP_SESSIONS
		if (sock->session)
			cp_hashlist_destroy_deep(sock->session);
		if (sock->pending)
			cp_hashlist_destroy_custom(sock->pending, NULL, 
					(cp_destructor_fn) cp_http_response_delete);
#endif /* USE_HTTP_SESSIONS */
		if (sock->service)
			cp_trie_destroy(sock->service);
		if (sock->server_name) 
			free(sock->server_name);
		free(sock);
	}
}

int cp_httpsocket_listen(cp_httpsocket *sock)
{
	int rc = cp_socket_listen(sock->sock);
	if (rc == 0) rc = cp_socket_select(sock->sock);

	return rc;
}

int cp_httpsocket_register_service(cp_httpsocket *server, 
		                           cp_http_service *service)
{
	int rc = 0;

	if (server->service == NULL)
	{
		server->service = 
			cp_trie_create_trie(COLLECTION_MODE_DEEP, NULL, 
								(cp_destructor_fn) cp_http_service_delete);
		if (server->service == NULL)
			return -1; //~~ error code
	}

	rc = cp_trie_add(server->service, service->path, service);

	return rc;
}

void *cp_httpsocket_unregister_service(cp_httpsocket *server, 
		                               cp_http_service *service)
{
	if (server->service)
	{
		void *rm = NULL;
		cp_trie_remove(server->service, service->path, &rm);
		return rm;
	}

	return NULL;
}


/****************************************************************************
 *                                                                          *
 * cp_http_service_* functions                                              *
 *                                                                          *
 ****************************************************************************/

cp_http_service *cp_http_service_create(char *name, 
                                        char *path, 
                                        cp_http_service_callback callback)
{
	cp_http_service *svc;
	
	svc = (cp_http_service *) calloc(1, sizeof(cp_http_service));
	if (svc == NULL)
	{
		cp_error(CP_MEMORY_ALLOCATION_FAILURE, "can\'t allocate cp_http_service");
		return NULL;
	}

	svc->name = strdup(name);
	svc->path = strdup(path);
	svc->service = callback;
	
	return svc;
}

void cp_http_service_delete(cp_http_service *service)
{
	if (service)
	{
		free(service->name);
		free(service->path);
		free(service);
	}
	else
		cp_warn("cp_http_service_delete: attempting to delete NULL service");
}


/****************************************************************************
 *                                                                          *
 * cp_http implementation helper functions                                  *
 *                                                                          *
 * o cp_http_default_response - creates a 501 NOT IMPLEMENTED response for  *
 *   the case that an cp_httpsocket instance is created with a NULL         *
 *   default_service                                                        *
 * o get_keepalive returns the keep-alive time for a request                *
 *                                                                          *
 ****************************************************************************/

static int cp_http_default_response(cp_http_request *request, cp_http_response *response)
{
	char buf[0x400];

	cp_warn("cp_http: no service defined for uri [%s]", request->uri);

	cp_http_response_set_content_type(response, HTML);
	cp_http_response_set_status(response, HTTP_501_NOT_IMPLEMENTED);

	sprintf(buf, "<html><head><title>%s: 501 NOT IMPLEMENTED</title></head><body><h1>Welcome to %s</h1>Your request [%s] could not be served: this server is not configured.<h1></body></html>", response->servername ? response->servername : DEFAULT_SERVER_NAME, response->servername ? response->servername : DEFAULT_SERVER_NAME, request->uri);

	response->body = strdup(buf);

	return HTTP_CONNECTION_POLICY_CLOSE;
}

static int get_keepalive(cp_httpsocket *sock, cp_http_request *request)
{
	int keepalive = 0;
	char *ka_lit;
	char *con_lit;
	
	con_lit = cp_hashtable_get(request->header, "Connection");
	if (con_lit && strcmp(con_lit, "close") == 0) return 0;
	
	ka_lit = cp_hashtable_get(request->header, "Keep-Alive");
	if (ka_lit)
		keepalive = atoi(ka_lit);
	
	if (keepalive > sock->keepalive || 
		(keepalive == 0 && request->version == HTTP_1_1)) 
		keepalive = sock->keepalive;

	return keepalive;
}

static int cp_httpselect(cp_httpsocket *hsock, cp_http_request *request, int fd)
{
	int rc = -1;
	int sec = get_keepalive(hsock, request);
#ifdef __TRACE__
	DEBUGMSG("keepalive: %d", sec);
#endif
	if (sec > 0)
	{
		fd_set rd;
		fd_set wr;
		fd_set err;
		struct timeval tv;
				
		FD_ZERO(&rd);
		FD_ZERO(&wr);
		FD_ZERO(&err);
		FD_SET(fd, &rd);
		FD_SET(fd, &wr);
		FD_SET(fd, &err);

		tv.tv_sec = sec;
		tv.tv_usec = 0;
//		rc = select(fd + 1, &rd, &wr, &err, &tv);
		rc = select(fd + 1, &rd, NULL, NULL, &tv);
#ifdef __TRACE__		
		DEBUGMSG("rd, wr, ex: %d, %d, %d",  
				FD_ISSET(fd, &rd), 
				FD_ISSET(fd, &wr),
				FD_ISSET(fd, &err));
#endif

		if (rc > -1) return FD_ISSET(fd, &rd) - 1;
#ifdef TRACE
		else DEBUGMSG("select: %s", strerror(errno));
#endif
			
	}
	
	return rc;
}

static int read_request(cp_connection_descriptor *cdesc, char *buf)
{
	char tmpbuf[HTTP_PARSE_BUFSIZE + 1];
	char *content = NULL;
	int body_len = 0;
	int body_read = -1;
	int rc = -1;
	int fd = cp_connection_descriptor_get_fd(cdesc);
#ifdef USE_SSL
	cp_socket *sock = cp_connection_descriptor_get_socket(cdesc);
#endif
	buf[0] = tmpbuf[0] = '\0';
	
	while (body_read < body_len)
	{
#ifdef USE_SSL
		if (sock->use_ssl)
		{
			if ((rc = SSL_read(cdesc->ssl, tmpbuf, HTTP_PARSE_BUFSIZE)) <= 0)
				break;
		}
		else
#endif
	    if ((rc = read(fd, tmpbuf, HTTP_PARSE_BUFSIZE)) <= 0) break;
	  
		tmpbuf[rc] = '\0';
		strcat(buf, tmpbuf);

		if (body_len == 0)
		{
			if (content == NULL) 
			{
				content = strstr(buf, "\r\n\r\n");
				if (content == NULL) 
					content = strstr(buf, "\n\r\n\r"); /* could happen */
				if (content) content += 2;
			}
			if (content == NULL) 
				content = strstr(buf, "\n\n");

			if (content)
			{
				char ch = *content;
				char *lenspec;
				int body_len = 0;

				*content = '\0';
				lenspec = strstr(buf, "Content-Length:");
				if (lenspec)
				{
					lenspec += 16;
					body_len = atoi(lenspec);
				}
				*content = ch;
				content += 2;
				body_read = strlen(content);
			}
		}
		else
			body_read += rc;
	}

	return rc;
}

/****************************************************************************
 *                                                                          *
 * cp_http_thread_fn - a cp_socket based partial http implementation        *
 *                                                                          *
 * the void *prm is a cp_socket cp_connection_descriptor. The               *
 * cp_socket->owner is a cp_httpsocket                                      *
 *                                                                          *
 ****************************************************************************/

void *cp_http_thread_fn(void *prm)
{
	cp_connection_descriptor *cdesc;
	cp_socket *sock;
	struct sockaddr_in *addr;
	int fd;
	char buf[HTTP_PARSE_BUFSIZE + 1]; 
	int rc = 0;
	int done;
	int action;
	cp_http_request *request = NULL; // prevents gcc warning
	cp_httpsocket *hsock;
	cp_http_service_callback svc;

	cdesc = (cp_connection_descriptor *) prm;
	sock = cp_connection_descriptor_get_socket(cdesc);
	fd = cp_connection_descriptor_get_fd(cdesc);
	hsock = (cp_httpsocket *) sock->owner;

	addr = cp_connection_descriptor_get_addr(cdesc);

#ifdef __TRACE__
	cp_info("\n\n *** connection from %s:%d\n", 
			inet_ntoa(addr->sin_addr), addr->sin_port);
#endif
	
	done = 0;

	while (!done && !sock->closing)
	{
		svc = hsock->default_service;
		action = HTTP_CONNECTION_POLICY_CLOSE;

		/* read client request */
		if ((rc = read_request(cdesc, buf)) == -1) /* io error */
		{
			rc = errno;
			cp_perror(CP_IO_ERROR, rc, "serving request from %s:%d on port %d",
					  inet_ntoa(addr->sin_addr), addr->sin_port,
					  sock->port);
			continue; //~~ check errno to decide whether to continue or break
		}

		/* handle empty requests or EOF */
		if (rc == 0 && buf[0] == '\0')  break;

		/* parse request and write response */
		if ((request = cp_http_request_parse(hsock, buf, &rc)))
		{
			cp_http_response *res;
#ifdef USE_HTTP_SESSIONS
			/* catch new session redirect and send pending responses */
			res = NULL;
			if (request->session && request->session->fresh)
			{
#ifdef USE_COOKIES
				if (request->session->type == SESSION_TYPE_COOKIE &&
					cp_http_request_get_parameter(request, CP_HTTP_SESSION_PRM))
					res = cp_http_new_session_redirect(request, 2);
				else
#endif /* USE_COOKIES */
				{
					res = cp_hashlist_get(hsock->pending, request->session->sid);
					if (res)
					{
						res->request = request;
						request->session->fresh = 0;
						cp_hashlist_remove(hsock->pending, request->session->sid);
					}
				}
			}

			if (res == NULL)
#endif /* USE_HTTP_SESSIONS */
			{
				res = cp_http_response_create(request);
				if (res == NULL)
				{
					cp_error(CP_MEMORY_ALLOCATION_FAILURE, "couldn\'t create response wrapper");
					cp_http_request_dump(request);
					cp_http_request_delete(request);
					continue;
				}

				/* determine service to invoke */
				if (hsock->service)
				{
					cp_http_service *desc;
					void *ptr;
					cp_trie_prefix_match(hsock->service, request->uri, &ptr);
					desc = ptr; /* prevents 'dereferencing type-punned pointer 
								   will break strict-aliasing rules' warning */
					if (desc) svc = desc->service;
#ifdef __TRACE__
					if (desc) DEBUGMSG("invoking service [%s]", desc->name);
#endif /* __TRACE__ */
				}
				if (svc == NULL)
					svc = cp_http_default_response;

				action = (*svc)(request, res);
				if (action != res->connection) //~~
					if (res->connection == HTTP_CONNECTION_POLICY_DEFAULT)
						res->connection = action;
				
#ifdef USE_HTTP_SESSIONS
				/* if service created a new session, start it */
				if (res->request->session)
				{ /* fresh session - rewrite url, set cookie, see what works */
					if (res->request->session->fresh)
					{
						if (hsock->pending == NULL)
							hsock->pending = 
								cp_hashlist_create(1, cp_hash_string, 
								                   cp_hash_compare_string);
						cp_hashlist_append(hsock->pending, 
										   res->request->session->sid, res);
						/* redirect to uri?cpsid=... */
						res = cp_http_new_session_redirect(request, 1);
					}
				}
#endif /* USE_HTTP_SESSIONS */
			}

			cp_http_response_write(cdesc, res);
			cp_http_response_delete(res);
		}
		else /* parse failed, rc should be set */
			cp_http_write_server_error(cdesc, rc);

		/* handle connection (keepalive or close as required) */
		switch (action)
		{
			case HTTP_CONNECTION_POLICY_KEEP_ALIVE: // select here
				if (cp_httpselect(hsock, request, fd) < 0)
					done = 1;
				break;

			case HTTP_CONNECTION_POLICY_CLOSE:
				done = 1;
				break;
		}
		if (request) cp_http_request_delete(request);
	}
			
#ifdef __TRACE__
	if (rc == 0)
		cp_info("done serving connection from %s:%d\n", 
				inet_ntoa(addr->sin_addr), addr->sin_port);
#endif

	cp_connection_descriptor_destroy(cdesc);

	return NULL;
}

/** @} */

