#ifndef _CP_HTTP_H
#define _CP_HTTP_H

/**
 * @addtogroup cp_socket
 */
/** @{ */
/**
 * @file
 * basic http feature implementations. the intention is to offer a simple way
 * of providing services to http based clients. 
 */

#include "common.h"

__BEGIN_DECLS

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "hashtable.h"
#include "trie.h"
#include "vector.h"
#include "socket.h"
#include "str.h"

#include <string.h>
#include <time.h>

/* HTTP 1.1 defines these methods (rfc 2616)
 * 
 *     Method         = "OPTIONS"
 *                    | "GET"   
 *                    | "HEAD"  
 *                    | "POST"  
 *                    | "PUT"   
 *                    | "DELETE"
 *                    | "TRACE" 
 *                    | "CONNECT"
 *
 *     general-header = Cache-Control     
 *                    | Connection        
 *                    | Date              
 *                    | Pragma            
 *                    | Trailer           
 *                    | Transfer-Encoding 
 *                    | Upgrade           
 *                    | Via               
 *                    | Warning           
 *
 *
 * 
 *     request-header = Accept             
 *                    | Accept-Charset     
 *                    | Accept-Encoding    
 *                    | Accept-Language    
 *                    | Authorization      
 *                    | Expect             
 *                    | From               
 *                    | Host               
 *                    | If-Match           
 *                    | If-Modified-Since  
 *                    | If-None-Match      
 *                    | If-Range           
 *                    | If-Unmodified-Since
 *                    | Max-Forwards       
 *                    | Proxy-Authorization
 *                    | Range              
 *                    | Referer            
 *                    | TE                 
 *                    | User-Agent         
 *
 *
 *     entity-header  = Allow              
 *                    | Content-Encoding   
 *                    | Content-Language   
 *                    | Content-Length     
 *                    | Content-Location   
 *                    | Content-MD5        
 *                    | Content-Range      
 *                    | Content-Type       
 *                    | Expires            
 *                    | Last-Modified      
 *
 *
 *     Status-Code    = "100"  : Continue
 *                    | "101"  : Switching Protocols
 *                    | "200"  : OK
 *                    | "201"  : Created
 *                    | "202"  : Accepted
 *                    | "203"  : Non-Authoritative Information
 *                    | "204"  : No Content
 *                    | "205"  : Reset Content
 *                    | "206"  : Partial Content
 *                    | "300"  : Multiple Choices
 *                    | "301"  : Moved Permanently
 *                    | "302"  : Found
 *                    | "303"  : See Other
 *                    | "304"  : Not Modified
 *                    | "305"  : Use Proxy
 *                    | "307"  : Temporary Redirect
 *                    | "400"  : Bad Request
 *                    | "401"  : Unauthorized
 *                    | "402"  : Payment Required
 *                    | "403"  : Forbidden
 *                    | "404"  : Not Found
 *                    | "405"  : Method Not Allowed
 *                    | "406"  : Not Acceptable
 *                    | "407"  : Proxy Authentication Required
 *                    | "408"  : Request Time-out
 *                    | "409"  : Conflict
 *                    | "410"  : Gone
 *                    | "411"  : Length Required
 *                    | "412"  : Precondition Failed
 *                    | "413"  : Request Entity Too Large
 *                    | "414"  : Request-URI Too Large
 *                    | "415"  : Unsupported Media Type
 *                    | "416"  : Requested range not satisfiable
 *                    | "417"  : Expectation Failed
 *                    | "500"  : Internal Server Error
 *                    | "501"  : Not Implemented
 *                    | "502"  : Bad Gateway
 *                    | "503"  : Service Unavailable
 *                    | "504"  : Gateway Time-out
 *                    | "505"  : HTTP Version not supported
 *
 * 
 *
 *    response-header = Accept-Ranges           ; Section 14.5
 *                    | Age                     ; Section 14.6
 *                    | ETag                    ; Section 14.19
 *                    | Location                ; Section 14.30
 *                    | Proxy-Authenticate      ; Section 14.33
 *                    | Retry-After             ; Section 14.37
 *                    | Server                  ; Section 14.38
 *                    | Vary                    ; Section 14.44
 *                    | WWW-Authenticate        ; Section 14.47
 *
 */
typedef enum 
{ 
	OPTIONS, 
	GET, 
	HEAD, 
	POST, 
	PUT, 
	DELETE, 
	TRACE, 
	CONNECT 
} cp_http_request_type;

typedef enum { HTTP_1_0, HTTP_1_1} cp_http_version;

typedef enum 
{
	HTTP_100_CONTINUE = 100,
	HTTP_101_SWITCHING_PROTOCOLS = 101,
	HTTP_200_OK = 200,
	HTTP_201_CREATED = 201,
	HTTP_202_ACCEPTED = 202,
	HTTP_203_NON_AUTHORITATIVE_INFORMATION = 203,
	HTTP_204_NO_CONTENT = 204,
	HTTP_205_RESET_CONTENT = 205,
	HTTP_206_PARTIAL_CONTENT = 206,
	HTTP_300_MULTIPLE_CHOICES = 300,
	HTTP_301_MOVED_PERMANENTLY = 301,
	HTTP_302_FOUND = 302,
	HTTP_303_SEE_OTHER = 303,
	HTTP_304_NOT_MODIFIED = 304,
	HTTP_305_USE_PROXY = 305,
	HTTP_307_TEMPORARY_REDIRECT = 307,
	HTTP_400_BAD_REQUEST = 400,
	HTTP_401_UNAUTHORIZED = 401,
	HTTP_402_PAYMENT_REQUIRED = 402,
	HTTP_403_FORBIDDEN = 403,
	HTTP_404_NOT_FOUND = 404,
	HTTP_405_METHOD_NOT_ALLOWED = 405,
	HTTP_406_NOT_ACCEPTABLE = 406,
	HTTP_407_PROXY_AUTHENTICATION_REQUIRED = 407,
	HTTP_408_REQUEST_TIME_OUT = 408,
	HTTP_409_CONFLICT = 409,
	HTTP_410_GONE = 410,
	HTTP_411_LENGTH_REQUIRED = 411,
	HTTP_412_PRECONDITION_FAILED = 412,
	HTTP_413_REQUEST_ENTITY_TOO_LARGE = 413,
	HTTP_414_REQUEST_URI_TOO_LARGE = 414,
	HTTP_415_UNSUPPORTED_MEDIA_TYPE = 415,
	HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE = 416,
	HTTP_417_EXPECTATION_FAILED = 417,
	HTTP_500_INTERNAL_SERVER_ERROR = 500,
	HTTP_501_NOT_IMPLEMENTED = 501,
	HTTP_502_BAD_GATEWAY = 502,
	HTTP_503_SERVICE_UNAVAILABLE = 503,
	HTTP_504_GATEWAY_TIME_OUT = 504,
	HTTP_505_HTTP_VERSION_NOT_SUPPORTED = 505
} cp_http_status_code;

typedef enum
{ 
	HTTP_CONNECTION_POLICY_DEFAULT,
	HTTP_CONNECTION_POLICY_CLOSE, 
  	HTTP_CONNECTION_POLICY_KEEP_ALIVE
} connection_policy;

typedef enum
{
	TEXT,
	HTML,
	JPEG
} cp_http_content_type;

#define DEFAULT_SERVER_NAME "libcprops-0.0.7"

#define DEFAULT_KEEPALIVE 300

#ifndef HTTP_KEEPALIVE
#define HTTP_KEEPALIVE DEFAULT_KEEPALIVE
#endif

#ifndef MAX_URL_LENGTH
#define MAX_URL_LENGTH   0x400
#endif

/** recommended to call before using cp_httpsocket api */
int cp_http_init();

/** call to perform cleanup after using cp_httpsocket api */
void cp_http_shutdown();

#ifdef USE_COOKIES

#ifndef MAX_COOKIE_LENGTH
#define MAX_COOKIE_LENGTH   0x1000
#endif /* MAX_COOKIE_LENGTH */

/* Wdy, DD-Mon-YYYY HH:MM:SS GMT */
#define COOKIE_TIME_FMT "%a, %d-%b-%Y %H:%M:%S GMT"
#endif /* USE_COOKIES */

#ifdef USE_HTTP_SESSIONS

#define CP_HTTP_SESSION_PRM "CPSID"
#define CP_HTTP_SESSION_MARKER "CPSID="
#define CP_HTTP_SESSION_MARKER_LEN 6

#define DEFAULT_SESSION_VALIDITY 86400 /* 24 hours */

/** 
 * call from signal handler to stop sockets in waiting select() and close all 
 * connections
 */
void cp_httpsocket_stop_all();

typedef enum 
{ 
	SESSION_TYPE_COOKIE = 1, 
	SESSION_TYPE_URLREWRITE 
} cp_http_session_type;

typedef struct _cp_http_session
{
	char *sid;
	cp_http_session_type type;
	time_t created;
	time_t access;
	long validity;
	cp_hashtable *key;
	short valid;
	short fresh; /**< newly created */
	int refcount;
} cp_http_session;

void *cp_http_session_get(cp_http_session *session, char *key);
inline void *cp_http_session_set(cp_http_session *session, 
                                 char *key, void *value);
void *cp_http_session_set_dtr(cp_http_session *session, 
		                      char *key, void *value, 
							  cp_destructor_fn dtr);
inline void cp_http_session_set_validity(cp_http_session *session, long sec);
inline short cp_http_session_is_fresh(cp_http_session *session);

#endif /* USE_HTTP_SESSIONS */


struct _cp_httpsocket;
		
/** http request descriptor */
typedef struct _cp_http_request
{
	cp_http_request_type type;
	cp_http_version version;
	char *uri;
	cp_hashtable *header;
	cp_hashtable *prm;
	cp_vector *prm_name;
#ifdef USE_COOKIES
	cp_vector *cookie;
#endif
	char *content;
	struct _cp_httpsocket *owner;
#ifdef USE_HTTP_SESSIONS
	cp_http_session *session;
#endif
} cp_http_request;

cp_http_request *cp_http_request_parse(struct _cp_httpsocket *owner, 
		                               char *request, 
                                       int *err);
void cp_http_request_delete(cp_http_request *request);
char *cp_http_request_get_header(cp_http_request *request, char *name);
char *cp_http_request_get_parameter(cp_http_request *request, char *name);
cp_vector *cp_http_request_get_param_vector(cp_http_request *request, char *name);
cp_vector *cp_http_request_get_params(cp_http_request *request);
#ifdef USE_HTTP_SESSIONS
cp_http_session *cp_http_request_get_session(cp_http_request *request, 
		                                     int create);
#endif
void cp_http_request_dump(cp_http_request *req);

/** http response holder */
typedef struct _cp_http_response
{
	cp_http_version version;
	cp_http_status_code status;
	cp_http_request *request;
	char *servername;
	connection_policy connection;
	cp_hashtable *header;
#ifdef USE_COOKIES
	cp_vector *cookie;
#endif
	cp_http_content_type content_type; /**< deprecated */
	char *content_type_lit;
	char *body; /**< deprecated */
	cp_string *content;
	int len;
} cp_http_response;

cp_http_response *cp_http_response_create(cp_http_request *request);
void cp_http_response_delete(cp_http_response *res);
int cp_http_response_write(cp_connection_descriptor *cdesc, 
						   cp_http_response *res);

inline void cp_http_response_set_status(cp_http_response *response, 
		                                cp_http_status_code code);
/** deprecated - use cp_http_response_set_content_type_string */
inline void cp_http_response_set_content_type(cp_http_response *response, 
		                                      cp_http_content_type type);
inline void cp_http_response_set_content_type_string(cp_http_response *response,
		                                             char *content_type_lit);
inline void cp_http_response_set_header(cp_http_response *response, 
		                                char *name, char *value);
/** deprecated - use cp_http_response_set_content instead */
inline void cp_http_response_set_body(cp_http_response *response, 
	                                  char *body);
inline void cp_http_response_set_content(cp_http_response *response, 
		                                 cp_string *content);

inline void 
	cp_http_response_set_connection_policy(cp_http_response *response, 
		                                   connection_policy policy);

#ifdef USE_COOKIES
int cp_http_response_set_cookie(cp_http_response *response, char *name, 
		char *content, char *host, char *path, long validity, int secure);
#endif

typedef int (*cp_http_service_callback)(cp_http_request *request, 
									 cp_http_response *response);

/** http server socket */
typedef struct _cp_httpsocket
{
	int id;
	cp_socket *sock;
	int keepalive; /* second resolution */
#ifdef USE_HTTP_SESSIONS
	cp_hashlist *session;
	cp_hashlist *pending;
#endif
	cp_trie *service; 
	cp_http_service_callback default_service;
	char *server_name;
} cp_httpsocket;


/** http service descriptor */
typedef struct _cp_http_service
{
	char *name;
	char *path;
	cp_http_service_callback service;
} cp_http_service;

void *cp_http_thread_fn(void *prm);

cp_httpsocket *
    cp_httpsocket_create(int port, cp_http_service_callback default_service);
#ifdef USE_SSL
cp_httpsocket *
    cp_httpsocket_create_ssl(int port, 
			                 cp_http_service_callback default_service,
							 char *certificate_file, 
							 char *key_file, 
							 int verification_mode);
#endif
void cp_httpsocket_delete(cp_httpsocket *svc);

inline void cp_httpsocket_set_keepalive(cp_httpsocket *socket, int sec);
inline void cp_httpsocket_set_server_name(cp_httpsocket *socket, char *name);
inline void cp_httpsocket_set_backlog(cp_httpsocket *socket, int backlog);
inline void cp_httpsocket_set_delay(cp_httpsocket *socket, struct timeval delay);
inline void cp_httpsocket_set_delay_sec(cp_httpsocket *socket, long sec);
inline void cp_httpsocket_set_delay_usec(cp_httpsocket *socket, long usec);
inline void cp_httpsocket_set_poolsize_min(cp_httpsocket *socket, int min);
inline void cp_httpsocket_set_poolsize_max(cp_httpsocket *socket, int max);

int cp_httpsocket_listen(cp_httpsocket *sock);

cp_http_service *cp_http_service_create(char *name, 
								  char *path, 
								  cp_http_service_callback service);
void cp_http_service_delete(cp_http_service *svc);

int cp_httpsocket_register_service(cp_httpsocket *server, cp_http_service *service);
void *cp_httpsocket_unregister_service(cp_httpsocket *server, cp_http_service *service);

__END_DECLS
/** @} */

#endif

