/**
 * @addtogroup cp_list
 */
/** @{ */
/**
 * @file
 * Implementation of cp_list Collection with linked elements and
 * cp_list_iterator.
 * 
 * The elements are stored in cp_list_entry objects.
 *
 * @copydoc collection
 */
/* ----------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
#include "linked_list.h"

/**
 * Creates an entry with the mode of the list.
 *
 * The new entry is not added to the list.
 * If the list has copy mode, it copies the item and puts the copy into the
 * new entry.
 *
 * @return entry object for the item.
 */
static cp_list_entry *cp_list_create_entry(cp_list *list, void *item);

//static void cp_list_destroy_entry(cp_list *list, cp_list_entry *entry);

/**
 * Unsynchronized and unchecked insert a new element at the beginning
 * of the list.
 */
static cp_list_entry *cp_list_insert_internal(cp_list *list, void *item);

/**
 * Unsynchronized and unchecked remove the entry from the list.
 */
static cp_list_entry *cp_list_remove_internal(cp_list *list, cp_list_entry *entry);

/**
 * Appends the element and returns the Entry.
 */
static cp_list_entry *cp_list_append_internal(cp_list *list, void *item);

/**
 * Removes the first entry from the list and returns it.
 *
 * Sets the references into a consistent state.
 */
static cp_list_entry *cp_list_remove_head_internal(cp_list *list);

/**
 * Removes the last entry from the list and returns it.
 *
 * Sets the references into a consistent state.
 */
static cp_list_entry *cp_list_remove_tail_internal(cp_list *list);


cp_list *cp_list_create_internal(int mode, 
								 cp_compare_fn compare_fn,
								 cp_copy_fn copy_fn,
								 cp_destructor_fn item_destructor,
								 int is_view)
{
    cp_list *list = NULL;

    list = (cp_list *) calloc(1, sizeof(cp_list));
    if (list == NULL) 
	{
        errno = ENOMEM;
        return NULL;
    }

    list->head       = NULL;
    list->tail       = NULL;
    list->compare_fn = compare_fn;
    list->copy_fn    = copy_fn;
	list->free_fn	 = item_destructor;
    list->mode       = mode;
    list->items      = 0;

	list->is_view    = is_view;

	if (!is_view)
	{
		list->lock = malloc(sizeof(pthread_rwlock_t));
		if (list->lock == NULL)
		{
			free(list);
			errno = ENOMEM;
			return NULL;
		}
	
	    pthread_rwlock_init(list->lock, NULL);
	}

    return list;
}

inline cp_list *cp_list_create()
{
	return 
		cp_list_create_internal(COLLECTION_MODE_MULTIPLE_VALUES,
				NULL, NULL, NULL, 0);
}

inline cp_list *cp_list_create_nosync()
{
	return 
		cp_list_create_internal(COLLECTION_MODE_MULTIPLE_VALUES |
							    COLLECTION_MODE_NOSYNC, 
								NULL, NULL, NULL, 0);
}

inline cp_list *cp_list_create_list(int mode, 
							 		cp_compare_fn compare_fn, 
							 		cp_copy_fn copy_fn,
								 	cp_destructor_fn item_destructor)
{
	return 
		cp_list_create_internal(mode, compare_fn, copy_fn, item_destructor, 0);
}


cp_list *cp_list_create_view(int mode, 
							 cp_compare_fn compare_fn, 
							 cp_copy_fn copy_fn,
							 cp_destructor_fn item_destructor,
							 cp_lock *lock)
{
	cp_list *list = 
		cp_list_create_internal(mode, compare_fn, copy_fn, item_destructor, 1);

	list->lock = lock; //~~ test

	return list;
}

void cp_list_destroy_internal(cp_list *list, cp_destructor_fn fn, int mode)
{
    cp_list_entry *curr, *next;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;

	if (!syncbit) cp_list_wrlock(list);
	
    curr = list->head;
    while (curr) 
	{
        next = curr->next;
        if ((mode & COLLECTION_MODE_DEEP) && fn)
			(*fn)(curr->item);
        free(curr);
        curr = next;
    }

	if (!syncbit)
	{
		cp_list_unlock(list);
		if (!list->is_view)
		{
			pthread_rwlock_destroy(list->lock);
			free(list->lock);
		}
	}

    free(list);
}

inline void cp_list_destroy(cp_list *list)
{
    cp_list_destroy_internal(list, list->free_fn, list->mode);
}

inline void cp_list_destroy_by_option(cp_list *list, int option)
{
	cp_list_destroy_internal(list, list->free_fn, option);
}

inline void cp_list_destroy_custom(cp_list *list, cp_destructor_fn fn)
{
    cp_list_destroy_internal(list, fn, list->mode | COLLECTION_MODE_DEEP);
}

inline long cp_list_item_count(cp_list *list)
{
    return (list == NULL) ? 0 : list->items;
}


static cp_list_entry **cp_list_get_entry_ref(cp_list *list, void *item)
{
    cp_list_entry **here = &list->head;

    while (*here != NULL && (*list->compare_fn)(item, (*here)->item)) 
        here = &(*here)-> next;

    return here;
}

void *cp_list_insert(cp_list *list, void *item)
{
    cp_list_entry *entry, **lookup;
    void *res = NULL;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;

    if (!syncbit) cp_list_wrlock(list);

    entry = NULL;
    if (!(list->mode & COLLECTION_MODE_MULTIPLE_VALUES)) 
	{
        lookup = cp_list_get_entry_ref(list, item);
        if (lookup) entry = *lookup;
    }

    if (entry == NULL) entry = cp_list_insert_internal(list, item);

    if (entry) res = entry->item;

    if (!syncbit) cp_list_unlock(list);
    
    return res;
}

void *cp_list_remove(cp_list *list, void *item)
{
    void *res = NULL;
    cp_list_entry *here, *curr;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;
	int mvalbit = list->mode & COLLECTION_MODE_MULTIPLE_VALUES;
	int deepbit = list->mode & COLLECTION_MODE_DEEP;

    if (!syncbit) cp_list_wrlock(list);

    here = list->head;
    while (here != NULL) 
	{
        curr = here;
        if ((*list->compare_fn)(item, curr->item) == 0) 
		{
            cp_list_remove_internal(list, curr);
            if (deepbit && list->free_fn) (*list->free_fn)(curr->item);
            free(curr);
    
            if (!mvalbit) break;
        }

        here = here->next;
    }

    if (!syncbit) cp_list_unlock(list);

    return res;
}

void *cp_list_insert_after(cp_list *list, void *item, void *existing)
{
    cp_list_entry **here, *entry;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;

    if (!syncbit) cp_list_wrlock(list);

    if (!(list->mode) & COLLECTION_MODE_MULTIPLE_VALUES) 
	{
        here = cp_list_get_entry_ref(list, item);
        if (*here != NULL) 
		{
            if (!syncbit) cp_list_unlock(list);
            return (*here)->item;
        }
    }

    entry = cp_list_create_entry(list, item);
    if (entry == NULL) 
	{
        if (!syncbit) cp_list_unlock(list);
        return NULL;
    }

    here = cp_list_get_entry_ref(list, existing);

    if (*here == NULL) /* no match - append to end of list */
		here = &list->tail;

	entry->prev = *here;
	entry->next = (*here)->next;
   	(*here)->next = entry;

	if (entry->next) 
		entry->next->prev = entry;
	else
		list->tail = entry;
    
    list->items++;
    
    if (!syncbit) cp_list_unlock(list);

    return entry->item;
}

void *cp_list_insert_before(cp_list *list, void *item, void *existing)
{
    cp_list_entry **here, *entry;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;

    if (!syncbit) cp_list_wrlock(list);

    if (!(list->mode) & COLLECTION_MODE_MULTIPLE_VALUES) 
	{
        here = cp_list_get_entry_ref(list, item);
        if (*here != NULL) 
		{
            if (!syncbit) cp_list_unlock(list);
            return (*here)->item;
        }
    }

    entry = cp_list_create_entry(list, item);
    if (entry == NULL) 
	{
        if (!syncbit) cp_list_unlock(list);
        return NULL;
    }

    here = cp_list_get_entry_ref(list, existing);

    if (*here == NULL) /* no match - insert at top of list */
		here = &list->head;
		
	entry->next = *here;
	entry->prev = (*here)->prev;
	(*here)->prev = entry;
	if (entry->prev)
		entry->prev->next = entry;
	else
		list->head = entry;

    list->items++;
        
    if (!syncbit) cp_list_unlock(list);

    return entry->item;
}

void *cp_list_search(cp_list *list, void *item)
{
    cp_list_entry **here;
    void *res;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;

    if (!syncbit) cp_list_rdlock(list);

    here = cp_list_get_entry_ref(list, item);
    res = *here ? (*here)->item : NULL;

    if (!syncbit) cp_list_unlock(list);
    
    return res;
}

int cp_list_callback(cp_list *l, int (*item_action)(void *, void *), void *id)
{
	int rc = 0;
	cp_list_entry *curr;
	int syncbit = l->mode & COLLECTION_MODE_NOSYNC;

	if (!syncbit) rc = cp_list_rdlock(l);

	if (rc) return rc;

	for (curr = l->head; curr; curr = curr->next)
		if ((rc = (*item_action)(curr->item, id)))
			break;

	if (!syncbit) cp_list_unlock(l);

	return rc;
}


void *cp_list_append(cp_list *list, void *item)
{
    cp_list_entry **lookup, *entry;
    void *res = NULL;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;

    if (!syncbit) cp_list_wrlock(list);

    if (!(list->mode & COLLECTION_MODE_MULTIPLE_VALUES)) 
	{
        lookup = cp_list_get_entry_ref(list, item);
        if (lookup != NULL) 
		{
            if (!syncbit) cp_list_unlock(list);
            return (*lookup)->item;
        }
    }

    entry = cp_list_append_internal(list, item);
    if (entry) res = entry->item;

    if (!syncbit) cp_list_unlock(list);

    return res;
}

void *cp_list_get_head(cp_list *list)
{
    void *item = NULL;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;
        
    if (!syncbit) cp_list_wrlock(list);

    if (list->head) item = list->head->item;

    if (!syncbit) cp_list_unlock(list);

    return item;
}

void *cp_list_get_tail(cp_list *list)
{
    void *item = NULL;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;
        
    if (!syncbit) cp_list_wrlock(list);

    if (list->tail) item = list->tail->item;

    if (!syncbit) cp_list_unlock(list);

    return item;
}


void *cp_list_remove_head(cp_list *list)
{
    cp_list_entry *old_head;
    void *res = NULL;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;
    
    if (!syncbit) cp_list_wrlock(list);

    old_head = cp_list_remove_head_internal(list);
    if (old_head) 
	{
        res = old_head->item;
		if ((list->mode & COLLECTION_MODE_DEEP) && list->free_fn)
			(*list->free_fn)(old_head->item);
		free(old_head);
    }

    if (!syncbit) cp_list_unlock(list);

    return res;
}

void *cp_list_remove_tail(cp_list *list)
{
    cp_list_entry *old_tail;
    void *res = NULL;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;

    if (!syncbit) cp_list_wrlock(list);

    old_tail = cp_list_remove_tail_internal(list);
    if (old_tail) 
	{
        res = old_tail->item;
        free(old_tail);
    }

    if (!syncbit) cp_list_unlock(list);

    return res;
}


int cp_list_is_empty(cp_list *list)
{
    int empty = 0;
	int syncbit = list->mode & COLLECTION_MODE_NOSYNC;

    if (!syncbit) cp_list_rdlock(list);

    empty = list->head == NULL;

    if (!syncbit) cp_list_unlock(list);

    return empty;
}

int cp_list_lock(cp_list *list, int mode)
{
    int rc;

    if (mode == COLLECTION_LOCK_READ) 
		rc = pthread_rwlock_rdlock(list->lock);
    else
		rc = pthread_rwlock_wrlock(list->lock);

    return rc;
}

int cp_list_unlock(cp_list *list)
{
    return pthread_rwlock_unlock(list->lock);
}

/****************************************************************************
 *                                                                          *
 *                    cp_list_iterator implementation                       *
 *                                                                          *
 ****************************************************************************/
 
cp_list_iterator* cp_list_create_iterator(cp_list *list, int type)
{
    int rc = - 1;
    cp_list_iterator *iterator = (cp_list_iterator *) malloc(sizeof(cp_list_iterator));
    iterator->list = list;
    iterator->pos = &list->head;
    iterator->lock_type = type;

    switch (type)
    {
        case COLLECTION_LOCK_READ : 
            rc = cp_list_rdlock(list);
            break;

        case COLLECTION_LOCK_WRITE :
            rc = cp_list_wrlock(list);
            break;
    
        default :
            rc = 0;
    }

	if (rc) /* locking failed */
	{
		free(iterator);
		iterator = NULL;
	}

    return iterator;
}

int cp_list_iterator_init(cp_list_iterator *iterator, cp_list *list, int type)
{
    int rc;
    iterator->list = list;
    iterator->pos = &list->head;
    iterator->lock_type = type;

    switch (type)
    {
        case COLLECTION_LOCK_READ : 
            rc = cp_list_rdlock(list);
            break;

        case COLLECTION_LOCK_WRITE :
            rc = cp_list_wrlock(list);
            break;
    
        default :
            rc = 0;
    }

    return rc;
}


int cp_list_iterator_head(cp_list_iterator *iterator)
{
    if (iterator == NULL) return -1;
    iterator->pos = &iterator->list->head;

    return 0;
}

int cp_list_iterator_tail(cp_list_iterator *iterator)
{
    if (iterator == NULL) return -1;
    iterator->pos = &iterator->list->tail;

    return 0;
}

int cp_list_iterator_init_tail(cp_list_iterator *iterator, cp_list *list, int type)
{
    int rc;
    iterator->list = list;
    iterator->pos = &list->tail;
    iterator->lock_type = type;

    switch (type)
    {
        case COLLECTION_LOCK_READ : 
            rc = cp_list_rdlock(list);
            break;

        case COLLECTION_LOCK_WRITE :
            rc = cp_list_wrlock(list);
            break;
    
        default :
            rc = 0;
    }

    return rc;
}

int cp_list_iterator_release(cp_list_iterator *iterator)
{
	int rc = 0;
    if (iterator->lock_type != COLLECTION_LOCK_NONE) 
		rc = cp_list_unlock(iterator->list);

    return rc;
}

int cp_list_iterator_destroy(cp_list_iterator *iterator)
{
    int rc = cp_list_iterator_release(iterator);
    free(iterator);

    return rc;
}

void *cp_list_iterator_next(cp_list_iterator *iterator)
{
    void *item = NULL;

    if (*(iterator->pos)) 
	{
        item = (*iterator->pos)->item;
        iterator->pos = &(*(iterator->pos))->next;
    }
	else if (iterator->list->head && 
			 iterator->pos == &iterator->list->head->prev)
	{
		item = iterator->list->head;
		iterator->pos = &iterator->list->head;
	}

    return item;
}

void *cp_list_iterator_prev(cp_list_iterator *iterator)
{
    void *item = NULL;

    if (*iterator->pos) 
	{
        item = (*iterator->pos)->item;
        iterator->pos = &(*iterator->pos)->prev;
    }
	else if (iterator->list->tail && 
			 iterator->pos == &iterator->list->tail->next)
	{
		item = iterator->list->tail;
		iterator->pos = &iterator->list->tail;
	}

    return item;
}

void *cp_list_iterator_curr(cp_list_iterator *iterator)
{
    void *item = NULL;

    if (*iterator->pos)
        item = (*iterator->pos)->item;

    return item;
}

void *cp_list_iterator_insert(cp_list_iterator *iterator, void *item)
{
	void *new_item = NULL;

	if ((iterator->list->mode & COLLECTION_MODE_NOSYNC) || 
		(iterator->lock_type == COLLECTION_LOCK_WRITE))
	{
		cp_list_entry *entry = cp_list_create_entry(iterator->list, item);
		if (entry == NULL) return NULL;
		new_item = entry->item;
		
		entry->next = *iterator->pos;

		if (*iterator->pos)
		{
			entry->prev = (*iterator->pos)->prev;
			(*iterator->pos)->prev = entry;
			if (entry->prev)
				entry->prev->next = entry;
		}
		else if (iterator->list->head == NULL) /* iterator not pointing at much - list may be empty */
			iterator->list->head = iterator->list->tail = entry;
		else if (iterator->pos == &iterator->list->head->prev) /* iterator moved before head */
		{
			entry->prev = NULL;
			entry->next = iterator->list->head;
			entry->next->prev = entry;
			iterator->list->head = entry;
		}
		else /* iterator moved after tail */
		{
			entry->prev = iterator->list->tail;
			entry->prev->next = entry;
			iterator->list->tail = entry;
		}

		iterator->pos =  &entry->next; /* keep iterator at same position */
		iterator->list->items++;
	}
	else /* mode is not NOSYNC and no LOCK_WRITE */
		errno = EINVAL;

	return new_item;
}

void *cp_list_iterator_append(cp_list_iterator *iterator, void *item)
{
	void *new_item = NULL;

	if ((iterator->list->mode & COLLECTION_MODE_NOSYNC) || 
		(iterator->lock_type == COLLECTION_LOCK_WRITE))
	{
		cp_list_entry *entry = cp_list_create_entry(iterator->list, item);
		if (entry == NULL) return NULL;
		new_item = entry->item;
		
		entry->prev = *iterator->pos;

		if (*iterator->pos)
		{
			entry->next = (*iterator->pos)->next;
			(*iterator->pos)->next = entry;
			if (entry->next)
				entry->next->prev = entry;
		}
		else if (iterator->list->tail == NULL) /* iterator not pointing at much - list may be empty */
			iterator->list->tail = iterator->list->head = entry;
		else if (iterator->pos == &iterator->list->tail->next) /* iterator moved after tail */
		{
			entry->next = NULL;
			entry->prev = iterator->list->tail;
			entry->prev->next = entry;
			iterator->list->tail = entry;
		}
		else /* iterator moved before head */
		{
			entry->next = iterator->list->head;
			entry->next->prev = entry;
			iterator->list->head = entry;
		}

		iterator->pos = &entry->prev; /* keep iterator at same position */
		iterator->list->items++;
	}
	else /* mode is not NOSYNC and no LOCK_WRITE */
		errno = EINVAL;

	return new_item;
}

void *cp_list_iterator_remove(cp_list_iterator *iterator)
{
	void *rm_item = NULL;

	if ((iterator->list->mode & COLLECTION_MODE_NOSYNC) || 
		(iterator->lock_type == COLLECTION_LOCK_WRITE))
	{
		if (*iterator->pos)
		{
			cp_list_entry *curr = *iterator->pos;
			if (curr->next)
				iterator->pos = &curr->next;
			else if (curr->prev)
				iterator->pos = &curr->prev;
			else /* removing last item */
				iterator->pos = &iterator->list->head;

			cp_list_remove_internal(iterator->list, curr);
		}
	}

	return rm_item;
}




/* ----------------------------------------------------------------- */
/** @} */

static cp_list_entry *cp_list_create_entry(cp_list *list, void *item)
{
    cp_list_entry *entry = (cp_list_entry *) calloc(1, sizeof(cp_list_entry));
    if (entry == NULL) 
	{
        errno = ENOMEM;
        return NULL;
    }
    entry->item = (list->mode & COLLECTION_MODE_COPY) ? (*list->copy_fn)(item) : item;

    return entry;
}

#if 0
static void cp_list_destroy_entry(cp_list *list, cp_list_entry *entry)
{
    if (list->mode & COLLECTION_MODE_DEEP) free(entry->item);

    free(entry);
}
#endif
 
static cp_list_entry *cp_list_insert_internal(cp_list *list, void *item)
{
    cp_list_entry *entry;

    entry = cp_list_create_entry(list, item);
    if (entry == NULL) return NULL;

    entry->next = list->head;
    list->head = entry;
    if (entry->next) entry->next->prev = entry;
    if (list->tail == NULL) list->tail = entry;

    list->items++;

    return entry;
}

static cp_list_entry *
	cp_list_remove_internal(cp_list *list, cp_list_entry *entry)
{
    if (entry->prev) 
		entry->prev->next = entry->next;
    else
		list->head = entry->next;

    if (entry->next)
		entry->next->prev = entry->prev;
    else
		list->tail = entry->prev;

    list->items--;

    return entry;
}

static cp_list_entry *cp_list_append_internal(cp_list *list, void *item)
{
    cp_list_entry *entry;

    entry = cp_list_create_entry(list, item);
    if (entry == NULL) return NULL;

    entry->prev = list->tail;
    list->tail = entry;

    if (entry->prev) entry->prev->next = entry;

    if (list->head == NULL) list->head = entry;

    list->items++;

    return entry;
}

static cp_list_entry *cp_list_remove_head_internal(cp_list *list)
{
    cp_list_entry *old_head;

    old_head = list->head;
    if (old_head) 
	{
        list->head = list->head->next;

        if (list->head == NULL) 
			list->tail = NULL;
        else
			list->head->prev = NULL;

        list->items--;
    }

    return old_head;
}

static cp_list_entry *cp_list_remove_tail_internal(cp_list *list)
{
    cp_list_entry *old_tail;

    old_tail = list->tail;
    if (old_tail) 
	{
        list->tail = list->tail->prev;

        if (list->tail == NULL) 
			list->head = NULL;
        else
			list->tail->next = NULL;

        list->items--;
    }

    return old_tail;
}

