#ifndef _CP_THREAD_H
#define _CP_THREAD_H

/**
 * @addtogroup cp_thread
 */
/** @{ */
/** 
 * @file
 * definitions for the cp_thread management framework. the
 * current implementation is based on the POSIX thread api. 
 */

#include "common.h"

__BEGIN_DECLS

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <pthread.h>
#include "linked_list.h"
#include "hashlist.h"
#include "vector.h"

#define cp_thread pthread_t
#define cp_thread_create(cp_thread, attr, fn, prm) pthread_create(&(cp_thread), attr, fn, prm)
#define cp_thread_join(cp_thread, exp) pthread_join(cp_thread, exp)
#define cp_thread_detach(cp_thread) pthread_detach(cp_thread)

/** a thread function */
typedef void *(*cp_thread_action) (void *);


/** a stop function for client threads */
typedef int (*cp_thread_stop_fn)(void *);

/**
 * cp_pooled_thread is a thread that lives in a cp_thread_pool. The struct 
 * holds synchronization elements used to control the thread and actuation 
 * settings, i.e. a thread function and a parameter to be passed for the next 
 * starting thread. The pool initializes a bunch of threads and sets them in 
 * 'wait'. 
 * When a client requests threads from the pool, the next available thread is
 * signalled out of 'wait', and runs the thread function requested by the 
 * client (see cp_thread_pool_get_impl). When the client thread function is 
 * done, the cp_pooled_thread returns to the thread pool and becomes available 
 * to pool clients. The cp_pooled_thread only exits when the pool exits, unless
 * explicitly stopped (eg pthread_exit) by client code. 
 */ 
typedef struct _cp_pooled_thread
{
	long id; ///< integral (implementation specific) cp_thread id
	struct _cp_thread_pool *owner; ///< the pool this cp_thread belongs to
	cp_thread *worker;	///< the actual cp_thread
	cp_thread_action action; ///< cp_thread function for the next assignment
	void *action_prm; ///< parameter for the next assignment
	cp_thread_stop_fn stop_fn; ///< called on cp_pooled_thread_stop
	void *stop_prm; ///< if not set, stop_fn invoked w/ action_prm

	pthread_mutex_t suspend_lock;	///< lock for framework scheduling
	pthread_cond_t suspend_cond;	///< condition variable for framework scheduling

	int done;	///< done flag
	int wait;	///< wait flag
} cp_pooled_thread;

/** return a cp_thread id for this cp_thread */
long cp_pooled_thread_get_id(cp_pooled_thread *cp_thread);

/** cp_thread constructor function */
cp_pooled_thread *cp_pooled_thread_create(struct _cp_thread_pool *owner);

/** cp_thread destructor */
void cp_pooled_thread_destroy(cp_pooled_thread *t);

/** signal a cp_thread to stop */
int cp_pooled_thread_stop(cp_pooled_thread *t);

/** retrieve an integer type cp_thread id for this cp_thread */
long cp_pooled_thread_get_id(cp_pooled_thread *t);

/** sets the action and prm for this cp_thread, then invokes 'action' */
int cp_pooled_thread_run_task(cp_pooled_thread *pt, 
							  cp_thread_action action, 
							  void *prm);

/** perform action with stop function */
int cp_pooled_thread_run_stoppable_task(cp_pooled_thread *pt, 
									  	cp_thread_action action,
									  	void *action_prm,
									  	cp_thread_stop_fn stop_fn,
									  	void *stop_prm);

/** framework cp_thread function */
void *cp_pooled_thread_run(void *prm);
	

/** 
 * cp_thread_pool holds a list of free cp_threads (in wait mode). The list grows up
 * to max_size, after which subsequent calls to cp_thread_pool_get will block, and
 * calls to cp_thread_pool_get_nb will return NULL - until clients return their 
 * cp_threads to the pool. 
 */
typedef struct _cp_thread_pool
{
	int size; ///< current size

	int min_size; ///< initial size
	int max_size; ///< size limit

	int running;

	pthread_mutex_t pool_lock; ///< to sync cp_thread assignment and release
	pthread_cond_t pool_cond; ///< to sync cp_thread assignment and release

	cp_list *free_pool; ///< holder for unused cp_threads
	cp_hashlist *in_use; ///< holder for running cp_threads
} cp_thread_pool;

/** cp_thread_pool constructor */
cp_thread_pool *cp_thread_pool_create(int min_size, int max_size);

/** cp_thread_pool destructor */
void cp_thread_pool_destroy(cp_thread_pool *pool);

/** signal all cp_threads in this pool to stop */
int cp_thread_pool_stop(cp_thread_pool *pool);

/** 
 * request a cp_thread from the pool. If no cp_threads are available, this function
 * will block until a cp_thread becomes available.
 */
inline cp_thread *cp_thread_pool_get(cp_thread_pool *pool, 
						   		  cp_thread_action action, 
						   		  void *prm);

/** 
 * request a cp_thread from the pool. If no cp_threads are available, this function
 * will block until a cp_thread becomes available.
 */
inline cp_thread *cp_thread_pool_get_stoppable(cp_thread_pool *pool, 
						   		  			cp_thread_action action, 
						   		  			void *action_prm, 
											cp_thread_stop_fn stop_fn,
											void *stop_prm);

/** 
 * request a cp_thread from the pool - non-blocking version. Returns a pointer 
 * to the requested cp_thread if one is available or NULL if the pool is empty. 
 */
inline cp_thread *cp_thread_pool_get_nb(cp_thread_pool *pool, 
							  	  cp_thread_action action, 
							  	  void *prm);
/** 
 * request a cp_thread from the pool - non-blocking version. Returns a pointer 
 * to the requested cp_thread if one is available or NULL if the pool is empty. 
 */
inline cp_thread *cp_thread_pool_get_stoppable_nb(cp_thread_pool *pool, 
							  	  cp_thread_action action, 
							  	  void *action_prm,
								  cp_thread_stop_fn stop_fn,
								  void *stop_prm);

/** returns the number of available cp_threads in the pool. */
inline int cp_thread_pool_count_available(cp_thread_pool *pool);

/* **************************************************************************
 *                                                                          *
 *                      cp_thread management framework                         *
 *                                                                          *
 ************************************************************************** */


/**
 * Definitions for cp_thread management framework follow. The framework is based
 * on the cp_thread_pool and cp_pooled_thread types. 
 * <br>
 * The pooled cp_thread scheduler interface is meant for use by clients who 
 * require a variable number of cp_threads. Each such component should create 
 * an instance of cp_pooled_thread_client_interface and use the api functions
 * to get cp_threads from the underlying cp_thread_pool. Here is some example
 * code. 
 * <p><tt><code><pre>
 * cp_pooled_thread_scheduler *main_scheduler; 
 * 
 * ...
 *
 * component_a_start(component_a *a, ...)
 * {
 *     a->scheduler_interface = 
 *         cp_pooled_thread_client_interface_create(main_scheduler, a, 2, 10, 
 *             component_a_report_load, component_a_stop_thread, 
 *             component_a_thread_run, a);
 * 
 *     ...
 * 
 *     for (i = 0; i < a->scheduler_interface->min; i++)
 *         cp_pooled_thread_client_get(a->scheduler_interface);
 * }
 *
 * component_b_start(component_b *b, ...)
 * {
 *     b->scheduler_interface = 
 *         cp_pooled_thread_client_interface_create(main_scheduler, b, 2, 10, 
 *             component_a_report_load, component_a_stop_thread, 
 *             component_a_thread_run, b);
 * 
 *     ...
 * 
 *     for (i = 0; i < b->scheduler_interface->min; i++)
 *         cp_pooled_thread_client_get(b->scheduler_interface);
 * }
 * 
 * </tt></code></pre><br>
 * In this example, the cp_threads for component_a and component_b will be 
 * managed jointly, since their cp_pooled_thread_client_interface *'s have the 
 * same cp_pooled_thread_scheduler *. <p>
 * 
 * 
 * See cp_pooled_thread_client_negociate for details. 
 */
typedef struct cp_pooled_thread_scheduler
{
	cp_thread_pool *pool; ///< pool to operate on
	cp_vector *client_list; ///< list of clients
	int bypass; ///< when bypass flag set 'negociate' function becomes nop
} cp_pooled_thread_scheduler;

struct _cp_pooled_thread_client_interface;

typedef int (*cp_pooled_thread_report_load) //~~ change prm to void *client
	(struct _cp_pooled_thread_client_interface *s);
typedef void (*cp_pooled_thread_shrink) //~~ change prm to void *client
	(struct _cp_pooled_thread_client_interface *);

/**
 * cp_pooled_thread_client_interface acts as the link to the 
 * cp_pooled_thread_scheduler for clients that require a variable number of 
 * cp_threads. This interface holds 3 functions pointers that must be supplied
 * by a client: <br>
 * <li> report_load - should return the number of open requests the client has
 * to handle
 * <li> shrink - will be called by the framework to stop one client cp_thread 
 * <li> action - the cp_thread function for this client
 */
typedef struct _cp_pooled_thread_client_interface
{
	int max;	///< cp_thread count upper limit 
	int min;	///< cp_thread count bottom limit

	int count;	///< actual number of cp_threads serving this client

	void *client;	///< pointer to client
	cp_pooled_thread_scheduler *owner;	///< pointer to cp_thread_pool_scheduler
	cp_pooled_thread_report_load report_load; ///< client function to report load
	cp_pooled_thread_shrink shrink; ///< client function to stop 1 cp_thread
	cp_thread_action action; ///< client cp_thread function 
	void *action_prm; ///< parameter to client cp_thread function (usually == client)
	cp_thread_stop_fn stop_fn; ///< stop callback 
	void *stop_prm; ///< parameter to stop callback - if NULL action_prm will be used
} cp_pooled_thread_client_interface;

/** cp_pooled_thread_client_interface constructor */
cp_pooled_thread_client_interface *
	cp_pooled_thread_client_interface_create
		(cp_pooled_thread_scheduler *owner, 
		 void *client, 
		 int min_threads, 
		 int max_threads,
		 cp_pooled_thread_report_load report_load,
		 cp_pooled_thread_shrink shrink,
		 cp_thread_action action,
		 void *action_prm,
		 cp_thread_stop_fn stop_fn,
		 void *stop_prm);

/** cp_pooled_thread_client_interface destructor */
void cp_pooled_thread_client_interface_destroy
	(cp_pooled_thread_client_interface *client);

/**
 * cp_threads should call negociate when a change in the number of cp_threads a 
 * client owns is required. 
 * Two possible scheduling approaches are -
 * 
 * (1) centralized:
 *     Clients report their load factor to the cp_thread manager. The cp_thread 
 *     manager grants requests for new cp_threads to clients with higher loads
 *     first. 
 * 
 * (2) distributed:
 *     clients negociate with other clients requesting a cp_thread based on their
 *     load factors. The client with the lower load factor releases a cp_thread 
 *     to the client with the higher load factor. 
 * 
 * The distributed approach saves some bookkeeping over head in the cp_thread 
 * manager, reduces the number steps involved in acquiring a new cp_thread or
 * releasing an unused one, and makes a dedicated synchronization cp_thread 
 * unnecessary. <p>
 * 
 * In the current implementation, the scheduler will randomly choose one 
 * other client to negociate with. If the load factors are different enough, 
 * one cp_thread will be switched to the busier client.
 */
void cp_pooled_thread_client_negociate(cp_pooled_thread_client_interface *c);

/** cp_pooled_thread_scheduler constructor */
cp_pooled_thread_scheduler *cp_pooled_thread_scheduler_create(cp_thread_pool *pool);

/** cp_pooled_thread_scheduler destructor */
void cp_pooled_thread_scheduler_destroy(cp_pooled_thread_scheduler *scheduler);

/** register client as a client of this scheduler */
void cp_pooled_thread_scheduler_register_client
		(cp_pooled_thread_scheduler *scheduler, 
		 cp_pooled_thread_client_interface *client);

/** 
 * convenience to abstract cp_thread_pool based implementation, see 
 * cp_pooled_thread_get and cp_pooled_thread_get_nb
 */
inline void cp_pooled_thread_client_get(cp_pooled_thread_client_interface *c);

/** 
 * convenience to abstract cp_thread_pool based implementation, see 
 * cp_pooled_thread_get and cp_pooled_thread_get_nb
 */
inline void cp_pooled_thread_client_get_stoppable(cp_pooled_thread_client_interface *c);

/** 
 * convenience to abstract cp_thread_pool based implementation, see 
 * cp_pooled_thread_get and cp_pooled_thread_get_nb
 */
inline void cp_pooled_thread_client_get_nb(cp_pooled_thread_client_interface *c);

/** 
 * convenience to abstract cp_thread_pool based implementation, see 
 * cp_pooled_thread_get and cp_pooled_thread_get_nb
 */
inline void cp_pooled_thread_client_get_stoppable_nb(cp_pooled_thread_client_interface *c);

__END_DECLS

/** @} */

#endif /* _CP_THREAD */

