#include <stdlib.h>
#include "collection.h"
#include "trie.h"
#include "mtab.h"

cp_trie_node *cp_trie_node_new(void *leaf) 
{ 
	cp_trie_node *node = calloc(1, sizeof(cp_trie_node)); 
	if (node) 
	{ 
		node->others = mtab_new(0); 
		node->leaf = leaf; 
	} 

	return node; 
} 

void *cp_trie_node_delete(cp_trie *grp, cp_trie_node *node) 
{ 
	void *leaf = NULL; 

	if (node) 
	{ 
		mtab_delete_custom(node->others, grp, (mtab_dtr)cp_trie_delete_mapping);
		leaf = node->leaf; 
		free(node); 
	} 

	if (leaf && grp->delete_leaf && grp->mode & COLLECTION_MODE_DEEP)
		(*grp->delete_leaf)(leaf);

	return leaf; 
} 

void cp_trie_delete_mapping(cp_trie *grp, mtab_node *map_node)
{
	if (map_node)
	{
		if (map_node->attr) free(map_node->attr);
		if (map_node->value) cp_trie_node_delete(grp, map_node->value);
	}
}

inline void cp_trie_node_unmap(cp_trie *grp, cp_trie_node **node) 
{ 
	cp_trie_node_delete(grp, *node); 
	*node = NULL; 
} 

int cp_trie_lock(cp_trie *grp, int type)
{
	int rc = 0;
	switch(type)
	{
		case COLLECTION_LOCK_READ:
			rc = pthread_rwlock_rdlock(grp->lock);
			break;

		case COLLECTION_LOCK_WRITE:
			rc = pthread_rwlock_wrlock(grp->lock);
			break;

		case COLLECTION_LOCK_NONE:
		default:
			break; //~~ warning?
	}

	return rc;
}

int cp_trie_unlock(cp_trie *grp)
{
	return pthread_rwlock_unlock(grp->lock);
}


cp_trie *cp_trie_create_trie(int mode,
		                     cp_copy_fn copy_leaf,
							 cp_destructor_fn delete_leaf) 
{ 
	cp_trie *grp = calloc(1, sizeof(cp_trie)); 

	if (grp == NULL) return NULL; 
	grp->root = cp_trie_node_new(NULL); 
	if (grp->root == NULL) 
	{ 
		free(grp); 
		return NULL; 
	} 

	if (!(mode & COLLECTION_MODE_NOSYNC))
	{
		if ((grp->lock = malloc(sizeof(pthread_rwlock_t))) == NULL)
		{
			cp_trie_node_delete(grp, grp->root);
			free(grp);
			return NULL;
		}

		if (pthread_rwlock_init(grp->lock, NULL)) 
		{ 
			free(grp->lock);
			cp_trie_node_delete(grp, grp->root);
			free(grp); 
			return NULL; 
		} 
	}

	grp->mode = mode;
	grp->copy_leaf = copy_leaf;
	grp->delete_leaf = delete_leaf;

	return grp; 
} 

inline cp_trie *cp_trie_create(mode)
{
	return cp_trie_create_trie(mode, NULL, NULL);
}

int cp_trie_destroy(cp_trie *grp) 
{ 
	int rc = 0; 

	if (grp)
	{ 
		pthread_rwlock_t *lock = grp->lock; 
		int mode = grp->mode; 

		if (!(mode & COLLECTION_MODE_NOSYNC)) 
		{ 
			if ((rc = cp_trie_wrlock(grp))) 
				return rc; 
		} 

		cp_trie_node_delete(grp, grp->root); 
		free(grp); 

		if (!(mode & COLLECTION_MODE_NOSYNC)) 
		{ 
			if ((rc = pthread_rwlock_unlock(lock))) 
				return rc; 
		} 

		if (lock) 
		{
			rc = pthread_rwlock_destroy(lock); 
			free(lock);
		}
	} 
	else rc = -1; 

	return rc; 
} 

void cp_trie_node_store(cp_trie *grp, 
                        cp_trie_node *node, 
						char *key, 
						void *leaf) 
{ 

	char *next;
	mtab_node *map_node;
	cp_trie_node *sub;

	map_node = NODE_MATCH(node, key); 
	if (map_node == NULL) /* not mapped - just add it */
	{ 
		sub = cp_trie_node_new(leaf); 
		NODE_STORE(node, key, sub); 
	} 
	else
	{
		next = map_node->attr;
		while (*key && *key == *next) 
		{
			key++;
			next++;
		}
		
		if (*next) /* branch abc, key abx or ab */
		{
			cp_trie_node *old = map_node->value;
			char *chopkey = strdup(next);
			sub = cp_trie_node_new(NULL);
			NODE_STORE(sub, chopkey, old);
			*next = '\0'; //~~ truncate and realloc - prevent dangling key
			map_node->value = sub;
			if (*key) /* key abx */
				NODE_STORE(sub, key, cp_trie_node_new(leaf));
			else /* key ab */
				sub->leaf = leaf;
		}
		else if (*key) /* branch abc, key abcde */
			cp_trie_node_store(grp, map_node->value, key, leaf);
		else  /* branch abc, key abc */
		{
			cp_trie_node *node = ((cp_trie_node *) map_node->value);
			if (node->leaf && grp->delete_leaf && 
				grp->mode & COLLECTION_MODE_DEEP)
				(*grp->delete_leaf)(node->leaf);
			node->leaf = leaf;
		}
	}
} 

int cp_trie_add(cp_trie *grp, char *key, void *leaf) 
{ 
	int rc = 0; 

	if (!(grp->mode & COLLECTION_MODE_NOSYNC)) 
	{ 
		if ((rc = cp_trie_wrlock(grp))) 
			return rc; 
	} 

	if (grp->mode & COLLECTION_MODE_COPY && grp->copy_leaf)
		leaf = (*grp->copy_leaf)(leaf);

	if (key == NULL) 
	{
		if (grp->root->leaf && 
				grp->delete_leaf && grp->mode & COLLECTION_MODE_DEEP)
			(*grp->delete_leaf)(grp->root->leaf);
		grp->root->leaf = leaf; 
	}
	else 
		cp_trie_node_store(grp, grp->root, key, leaf); 

	grp->path_count++; 

	if (!(grp->mode & COLLECTION_MODE_NOSYNC)) 
	{ 
		if ((rc = cp_trie_unlock(grp))) 
			return rc; 
	} 

	return rc; 
} 

int cp_trie_remove(cp_trie *grp, char *key, void **leaf) 
{ 
	int rc = 0; 
	char *i = key; 
	cp_trie_node **link = &grp->root; 
	mtab_node *map_node;
	cp_trie_node *next; 
	cp_trie_node **last = link; 

	if (!(grp->mode & COLLECTION_MODE_NOSYNC)) 
	{ 
		if ((rc = cp_trie_wrlock(grp))) 
			return rc; 
	} 

	while (i && (*link) && (map_node = NODE_MATCH(*link, i))) 
	{
		char *branch = map_node->attr;
		while (*i && *branch == *i) //~~
		{
			branch++;
			i++;
		}
		if (*branch) break; /* mismatch - not found */

		next = map_node->value;
		if ((*link)->leaf || BRANCH_COUNT(*link) > 1) *last = *link; 
		*link = next; 
	} 

	if (*last) 
	{ 
		grp->path_count--; 
		cp_trie_node_unmap(grp, last); 
		*leaf = (*link)->leaf; 
	} 
	else 
		*leaf = NULL; 
	
	if (!(grp->mode & COLLECTION_MODE_NOSYNC)) 
	{ 
		if ((rc = cp_trie_unlock(grp))) 
			return rc; 
	} 
 
	return rc; 
} 

void *cp_trie_exact_match(cp_trie *grp, char *key)
{
	void *last = NULL;
	cp_trie_node *link = grp->root; 
	mtab_node *map_node;
	char *branch;

	if (!(grp->mode & COLLECTION_MODE_NOSYNC)) 
	{ 
		if (cp_trie_rdlock(grp))
			return NULL; 
	} 
 
	while ((map_node = NODE_MATCH(link, key)) != NULL) 
	{
		branch = map_node->attr;
 
		while (*key && *key == *branch) //~~
		{
			key++;
			branch++;
		}
		if (*branch) break; /* mismatch */

		link = map_node->value; 
	} 

	if (link) last = link->leaf;

	if (!(grp->mode & COLLECTION_MODE_NOSYNC)) 
	{ 
		if (cp_trie_unlock(grp))
			return NULL; 
	} 
 
	if (*key == '\0') return last;  // Avoid match on super-string keys
	else return NULL;
}

int cp_trie_prefix_match(cp_trie *grp, char *key, void **leaf)
{ 
	int rc, match_count = 0; 
	void *last = grp->root->leaf; 
	cp_trie_node *link = grp->root; 
	mtab_node *map_node;
	char *branch;

	if (!(grp->mode & COLLECTION_MODE_NOSYNC)) 
	{ 
		if ((rc = cp_trie_rdlock(grp))) 
			return rc; 
	} 
 
	while ((map_node = NODE_MATCH(link, key)) != NULL) 
	{ 
		branch = map_node->attr;
 
		while (*key && *key == *branch)
		{
			key++;
			branch++;
		}
		if (*branch) break; /* mismatch */

		link = map_node->value; 
		if (link->leaf)
		{ 
			last = link->leaf; 
			match_count++; 
		} 
	} 

	*leaf = last; 
 
	if (!(grp->mode & COLLECTION_MODE_NOSYNC)) 
	{ 
		if ((rc = cp_trie_unlock(grp))) 
			return rc; 
	} 
 
	return match_count; 
} 

inline void cp_trie_set_root(cp_trie *grp, void *leaf) 
{ 
    grp->root->leaf = leaf; 
} 

static int node_count;
static int depth_total;
static int max_level;

void cp_trie_dump_node(cp_trie_node *node, int level, char *prefix)
{
	int i;
	mtab_node *map_node;

	node_count++;
	depth_total += level;
	if (level > max_level) max_level = level;

	for (i = 0; i < node->others->size; i++)
	{
		map_node = node->others->table[i];
		while (map_node)
		{
			cp_trie_dump_node(map_node->value, level + 1, map_node->attr);
			map_node = map_node->next;
		}
	}

	for (i = 0; i < level; i++) printf("\t");

	printf(" - %s => [%s]\n", prefix, node->leaf ? (char *) node->leaf : "");
}
			
void cp_trie_dump(cp_trie *grp)
{
	node_count = 0;
	depth_total = 0;
	max_level = 0;
	
	cp_trie_dump_node(grp->root, 0, "");

	printf("\n %d nodes, %d deepest, avg. depth is %.2f\n\n", 
			node_count, max_level, (float) depth_total / node_count);
}

