/**
 * cp_vector is a 'safe array' implementation
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "log.h"
#include "common.h"

#include "vector.h"

cp_vector *cp_vector_create_by_option(int size, 
		                              int item_size, 
									  int mode, 
									  cp_copy_fn copy_item,
									  cp_destructor_fn free_item)
{
	cp_vector *v = calloc(1, sizeof(cp_vector));
	if (v == NULL) 
	{
		errno = ENOMEM;
		return NULL;
	}

	v->mem = calloc(size, sizeof(void *));
	if (v->mem == NULL)
	{
		errno = ENOMEM;
		return NULL;
	}

	v->size = size;
	v->item_size = item_size;
	v->mode = mode;
	v->copy_item = copy_item;
	v->free_item = free_item;
	v->head = v->tail = 0;

	return v;
}

inline cp_vector *cp_vector_create(int size, int item_size)
{
	return cp_vector_create_by_option(size, item_size, 0, NULL, NULL);
}

void cp_vector_destroy(cp_vector *v)
{
	free(v->mem);
	free(v);
}

void cp_vector_destroy_custom(cp_vector *v, cp_destructor_fn dtr)
{
	int i;
	int n = cp_vector_size(v);

	if (dtr)
	{
		for (i = 0; i < n; i++)
			if (v->mem[i]) (*dtr)(v->mem[i]);
	}
	
	cp_vector_destroy(v);
}

void *cp_vector_element_at(cp_vector *v, int index)
{
	return v->mem[index];
}

#if 0
void *cp_vector_add_element_at(cp_vector *v, int index, void *element)
{
	if (v->head - v->tail + 1 >= v->size)
	{
		v->mem = realloc(v->mem, v->size * v->item_size * 2);
		
		if (v->mem == NULL)
		{
			errno = ENOMEM;
			return NULL;
		}
		v->size *= 2;
	}

	if (index <= v->head - v->tail)
	{
		memmove(v->mem + v->item_size * v->tail, 
				v->mem + v->item_size * (v->tail + 1),
				v->item_size * (v->head + 1 - v->tail));
		memcpy(v->mem + v->tail * v->item_size, element, v->item_size);
		v->tail++;
	}
	else 
	{
		
	}
}
#endif

void *cp_vector_add_element(cp_vector *v, void *element)
{
	void *addr = NULL;

	if (v->head + 1 >= v->tail + v->size)
	{
		v->mem = realloc(v->mem, 2 * v->size * sizeof(void *));
		if (v->head < v->tail)
		{
			memcpy(v->mem, &v->mem[v->size], v->head * sizeof(void *));
			v->head += v->size;
		}
		v->size *= 2;
	}
			
	if (v->mode & COLLECTION_MODE_COPY)
		v->mem[v->head] = (*v->copy_item)(element);
	else
		v->mem[v->head] = element;
	addr = v->mem[v->head];
	v->head = (v->head + 1) % v->size;

	return addr;
#if 0
	if (v->head - v->tail + 1 >= v->size) /* full, realloc */
	{
		v->mem = realloc(v->mem, v->size * 2);
		if (v->mem == NULL)
		{
			errno = ENOMEM;
			return NULL;
		}
		v->size *= 2;
	}

	if (v->head + 1 >= v->size)
	{
		if (v->tail > 0)
		{
			v->tail--;
			addr = v->mem + v->tail * v->item_size;
			memcpy(addr, element, v->item_size);
		}
		else cp_fatal(-1, "cp_vector: implementation error"); //~~ debug
	}
	else
	{
		addr = v->mem + v->head * v->item_size;
		memcpy(addr, element, v->item_size);
		v->head++;
	}

	return addr;
#endif
}

int cp_vector_size(cp_vector *v)
{
	return (v->head - v->tail + v->size) % v->size;
}

